# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from astropy.table import Column, Table
from gammapy.data import DataStore

__all__ = ["to_obscore_table"]


log = logging.getLogger(__name__)

DEFAULT_OBSCORE_TEMPLATE = {
    "dataproduct_type": "event",
    "calib_level": 2,
    "obs_collection": "DL3",
    "access_format": "application/fits",
    "s_fov": 10.0,
}


def empty_obscore_table():
    """Generate the Obscore default table.

    In case the obscore standard changes, this function should be changed according
    to https://www.ivoa.net/documents/ObsCore

    Returns
    -------
    table : `~astropy.table.Table`
       the empty table
    """
    obscore_table = [None] * 29
    obscore_table[0] = Column(
        name="dataproduct_type",
        unit="",
        description="Data product (file content) primary type",
        dtype="U10",
        meta={"Utype": "ObsDataset.dataProductType", "UCD": "meta.id"},
    )
    obscore_table[1] = Column(
        name="calib_level",
        unit="",
        description="Calibration level of the observation: in {0, 1, 2, 3, 4}",
        dtype="i4",
        meta={"Utype": "ObsDataset.calibLevel", "UCD": "meta.code;obs.calib"},
    )
    obscore_table[2] = Column(
        name="target_name",
        unit="",
        description="Object of interest",
        dtype="U25",
        meta={"Utype": "Target.name", "UCD": "meta.id;src"},
    )
    obscore_table[3] = Column(
        name="obs_id",
        unit="",
        description="Internal ID given by the ObsTAP service",
        dtype="U10",
        meta={"Utype": "DataID.observationID", "UCD": "meta.id"},
    )
    obscore_table[4] = Column(
        name="obs_collection",
        unit="",
        description="Name of the data collection",
        dtype="U10",
        meta={"Utype": "DataID.collection", "UCD": "meta.id"},
    )
    obscore_table[5] = Column(
        name="obs_publisher_did",
        unit="",
        description="ID for the Dataset given by the publisher",
        dtype="U30",
        meta={"Utype": "Curation.publisherDID", "UCD": "meta.ref.uri;meta.curation"},
    )
    obscore_table[6] = Column(
        name="access_url",
        unit="",
        description="URL used to access dataset",
        dtype="U30",
        meta={"Utype": "Access.reference", "UCD": "meta.ref.url"},
    )
    obscore_table[7] = Column(
        name="access_format",
        unit="",
        description="Content format of the dataset",
        dtype="U30",
        meta={"Utype": "Access.format", "UCD": "meta.code.mime"},
    )
    obscore_table[8] = Column(
        name="access_estsize",
        unit="kbyte",
        description="Estimated size of dataset: in kilobytes",
        dtype="i4",
        meta={"Utype": "Access.size", "UCD": "phys.size;meta.file"},
    )
    obscore_table[9] = Column(
        name="s_ra",
        unit="deg",
        description="Central Spatial Position in ICRS Right ascension",
        dtype="f8",
        meta={
            "Utype": "Char.SpatialAxis.Coverage.Location.Coord.Position2D.Value2.C1",
            "UCD": "pos.eq.ra",
        },
    )
    obscore_table[10] = Column(
        name="s_dec",
        unit="deg",
        description="Central Spatial Position in ICRS Declination",
        dtype="f8",
        meta={
            "Utype": "Char.SpatialAxis.Coverage.Location.Coord.Position2D.Value2.C2",
            "UCD": "pos.eq.dec",
        },
    )
    obscore_table[11] = Column(
        name="s_fov",
        unit="deg",
        description="Estimated size of the covered region as the diameter of a containing circle",
        dtype="f8",
        meta={
            "Utype": "Char.SpatialAxis.Coverage.Bounds.Extent.diameter",
            "UCD": "phys.angSize;instr.fov",
        },
    )
    obscore_table[12] = Column(
        name="s_region",
        unit="",
        description="Sky region covered by the data product (expressed in ICRS frame)",
        dtype="U30",
        meta={
            "Utype": "Char.SpatialAxis.Coverage.Support.Area",
            "UCD": "pos.outline;obs.field",
        },
    )
    obscore_table[13] = Column(
        name="s_resolution",
        unit="arcsec",
        description="Spatial resolution of data as FWHM of PSF",
        dtype="f8",
        meta={
            "Utype": "Char.SpatialAxis.Resolution.refval.value",
            "UCD": "pos.angResolution",
        },
    )
    obscore_table[14] = Column(
        name="s_xel1",
        unit="",
        description="Number of elements along the first coordinate of the spatial axis",
        dtype="i4",
        meta={"Utype": "Char.SpatialAxis.numBins1", "UCD": "meta.number"},
    )
    obscore_table[15] = Column(
        name="s_xel2",
        unit="",
        description="Number of elements along the second coordinate of the spatial axis",
        dtype="i4",
        meta={"Utype": "Char.SpatialAxis.numBins2", "UCD": "meta.number"},
    )
    obscore_table[16] = Column(
        name="t_xel",
        unit="",
        description="Number of elements along the time axis",
        dtype="i4",
        meta={"Utype": "Char.TimeAxis.numBins", "UCD": "meta.number"},
    )
    obscore_table[17] = Column(
        name="t_min",
        unit="d",
        description="Start time in MJD",
        dtype="f8",
        meta={
            "Utype": "Char.TimeAxis.Coverage.Bounds.Limits.StartTime",
            "UCD": "time.start;obs.exposure",
        },
    )
    obscore_table[18] = Column(
        name="t_max",
        unit="d",
        description="Stop time in MJD",
        dtype="f8",
        meta={
            "Utype": "Char.TimeAxis.Coverage.Bounds.Limits.StopTime",
            "UCD": "time.end;obs.exposure",
        },
    )
    obscore_table[19] = Column(
        name="t_exptime",
        unit="s",
        description="Total exposure time",
        dtype="f8",
        meta={
            "Utype": "Char.TimeAxis.Coverage.Support.Extent",
            "UCD": "time.duration;obs.exposure",
        },
    )
    obscore_table[20] = Column(
        name="t_resolution",
        unit="s",
        description="Temporal resolution FWHM",
        dtype="f8",
        meta={
            "Utype": "Char.TimeAxis.Resolution.Refval.valueResolution.Refval.value",
            "UCD": "time.resolution",
        },
    )
    obscore_table[21] = Column(
        name="em_xel",
        unit="",
        description="Number of elements along the spectral axis",
        dtype="i4",
        meta={"Utype": "Char.SpectralAxis. numBins", "UCD": "meta.number"},
    )
    obscore_table[22] = Column(
        name="em_min",
        unit="TeV",
        description="start in spectral coordinates",
        dtype="f8",
        meta={
            "Utype": "Char.SpectralAxis.Coverage.Bounds.Limits.LoLimit",
            "UCD": "em.wl;stat.min",
        },
    )
    obscore_table[23] = Column(
        name="em_max",
        unit="TeV",
        description="stop in spectral coordinates",
        dtype="f8",
        meta={
            "Utype": "Char.SpectralAxis.Coverage.Bounds.Limits.HiLimit",
            "UCD": "em.wl;stat.max",
        },
    )
    obscore_table[24] = Column(
        name="em_res_power",
        unit="",
        description="Value of the resolving power along the spectral axis(R)",
        dtype="f8",
        meta={
            "Utype": "Char.SpectralAxis.Resolution.ResolPower.refVal",
            "UCD": "spect.resolution",
        },
    )
    obscore_table[25] = Column(
        name="o_ucd",
        unit="",
        description="Nature of the observable axis",
        dtype="U30",
        meta={"Utype": "Char.ObservableAxis.ucd", "UCD": "meta.ucd"},
    )
    obscore_table[26] = Column(
        name="pol_xel",
        unit="",
        description="Number of elements along the polarization axis",
        dtype="i4",
        meta={"Utype": "Char.PolarizationAxis.numBins", "UCD": "meta.number"},
    )
    obscore_table[27] = Column(
        name="facility_name",
        unit="",
        description="The name of the facility, telescope space craft used for the observation",
        dtype="U10",
        meta={
            "Utype": "Provenance.ObsConfig.Facility.name",
            "UCD": "meta.id;instr.tel",
        },
    )
    obscore_table[28] = Column(
        name="instrument_name",
        unit="",
        description="The name of the instrument used for the observation",
        dtype="U25",
        meta={"Utype": "Provenance.ObsConfig.Instrument.name", "UCD": "meta.id;instr"},
    )
    return Table(obscore_table)


def observation_obscore_dict(observation):
    """Generate an obscore dictionary from an observation.

    Parameters
    ----------
    observation : `~gammapy.data.Observation`
        the observation

    Returns
    -------
    result : dict
    """
    return {
        "target_name": observation.meta.target.name,
        "obs_id": str(observation.obs_id),
        "s_ra": observation.get_pointing_icrs(observation.tmid).ra.to_value("deg"),
        "s_dec": observation.get_pointing_icrs(observation.tmid).dec.to_value("deg"),
        "t_min": observation.tstart.to_value("mjd"),
        "t_max": observation.tstop.to_value("mjd"),
        "t_exptime": observation.observation_live_time_duration.to_value("s"),
        "em_min": observation.events.energy.min().value,
        "em_max": observation.events.energy.max().value,
        "facility_name": observation.meta.obs_info.telescope,
        "instrument_name": observation.meta.obs_info.instrument,
    }


def to_obscore_table(
    base_dir,
    selected_obs=None,
    obs_publisher_did=None,
    access_url=None,
    obscore_template=None,
):
    """Generate the complete obscore Table by adding one row per observation using _obscore_row().

    Parameters
    ----------
    base_dir : str or `~pathlib.Path`
        Base directory of the data files.
    selected_obs : list or array of Observation ID(int)
        Default is None (default of ``None`` means ``no observation ``).
        If not given, the full obscore (for all the obs_ids in DataStore) table is returned.
    obs_publisher_did : str, optional
        ID for the Dataset given by the publisher (check IVOA recommendations).
        Default is None. Giving the values of this argument is highly recommended.
        If not the corresponding obscore field is filled by the Observation ID value.
    access_url : str, optional
        URL used to to access (download) dataset(check IVOA recommendations).
        Default is None. Giving the values of this argument is highly recommended.
        If not the corresponding obscore field is filled by the Observation ID value.
    obscore_template : dict, optional
        Template for fixed values in the obscore Table.
        Default is DEFAULT_OBSCORE_TEMPLATE

    Returns
    -------
    obscore_tab : ~astropy.table.Table
        Obscore table with number of rows = len(selected_obs)

    References
    ----------
        * `IVOA ObsCore recommendations <https://www.ivoa.net/documents/ObsCore>`_
        * `https://www.ivoa.net/documents/ObsCore and https://www.ivoa.net/documents/TAP/`_
        * `IVOA identifiers <https://www.ivoa.net/documents/IVOAIdentifiers/20160523/REC-Identifiers-2.0.html>`_
    """
    if obs_publisher_did is None:
        log.warning(
            "Insufficient publisher information: 'obs_publisher_did'. Giving this value is highly recommended."
        )
        obs_publisher_did = ""
    if access_url is None:
        log.warning(
            "Insufficient publisher information: 'access_url'. Giving this value is highly recommended."
        )
        access_url = ""

    if obscore_template is None:
        log.info("No template provided, using DEFAULT_OBSCORE_TEMPLATE")

    result = DEFAULT_OBSCORE_TEMPLATE.copy()
    if obscore_template is not None:
        result.update(obscore_template)

    data_store = DataStore.from_dir(base_dir)
    if selected_obs is None:
        selected_obs = data_store.obs_ids

    obscore_table = empty_obscore_table()

    for obs_id in selected_obs:
        obscore_row = result.copy()

        hdu_loc = data_store.hdu_table.hdu_location(obs_id, "events")
        obscore_row["obs_publisher_did"] = (f"{obs_publisher_did}#{obs_id}",)
        obscore_row["access_estsize"] = hdu_loc.path().stat().st_size / 1024.0
        obscore_row["access_url"] = f"{access_url}{hdu_loc.file_name}"

        observation = data_store.obs(obs_id)
        obscore_row.update(observation_obscore_dict(observation))

        obscore_table.add_row(obscore_row)
    return obscore_table



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Supernova remnant (SNR) source models."""
import html
import numpy as np
import astropy.constants
from astropy.units import Quantity
from astropy.utils import lazyproperty

__all__ = ["SNR", "SNRTrueloveMcKee"]


class SNR:
    """Simple supernova remnant (SNR) evolution model.

    The model is based on the Sedov-Taylor solution for strong explosions.

    Reference: https://ui.adsabs.harvard.edu/abs/1950RSPSA.201..159T

    Parameters
    ----------
    e_sn : `~astropy.units.Quantity`
        SNR energy (erg), equal to the SN energy after neutrino losses.
    theta : `~astropy.units.Quantity`
        Fraction of E_SN that goes into cosmic rays.
    n_ISM : `~astropy.units.Quantity`
        ISM density (g cm^-3).
    m_ejecta : `~astropy.units.Quantity`
        Ejecta mass (g).
    t_stop : `~astropy.units.Quantity`
        Post-shock temperature where gamma-ray emission stops.
    """

    def __init__(
        self,
        e_sn="1e51 erg",
        theta=Quantity(0.1),
        n_ISM=Quantity(1, "cm-3"),
        m_ejecta=astropy.constants.M_sun,
        t_stop=Quantity(1e6, "K"),
        age=None,
        morphology="Shell2D",
        spectral_index=2.1,
    ):
        self.e_sn = Quantity(e_sn, "erg")
        self.theta = theta
        self.rho_ISM = n_ISM * astropy.constants.m_p
        self.n_ISM = n_ISM
        self.m_ejecta = m_ejecta
        self.t_stop = t_stop
        self.morphology = morphology
        self.spectral_index = spectral_index
        if age is not None:
            self.age = Quantity(age, "yr")

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def radius(self, t):
        r"""Outer shell radius at age t.

        The radius during the free expansion phase is given by:

        .. math::
            r_{SNR}(t) \approx 0.01
            \left(\frac{E_{SN}}{10^{51}erg}\right)^{1/2}
            \left(\frac{M_{ej}}{M_{\odot}}\right)^{-1/2} t
            \text{ pc}

        The radius during the Sedov-Taylor phase evolves like:

        .. math::
            r_{SNR}(t) \approx \left(\frac{E_{SN}}{\rho_{ISM}}\right)^{1/5}t^{2/5}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        t = Quantity(t, "yr")
        r = np.where(
            t > self.sedov_taylor_begin,
            self._radius_sedov_taylor(t).to_value("cm"),
            self._radius_free_expansion(t).to_value("cm"),
        )
        return Quantity(r, "cm")

    def _radius_free_expansion(self, t):
        """Shock radius at age t during free expansion phase.

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        # proportional constant for the free expansion phase
        term_1 = (self.e_sn / Quantity(1e51, "erg")) ** (1.0 / 2)
        term_2 = (self.m_ejecta / astropy.constants.M_sun) ** (-1.0 / 2)
        return Quantity(0.01, "pc/yr") * term_1 * term_2 * t

    def _radius_sedov_taylor(self, t):
        """Shock radius  at age t  during Sedov Taylor phase.

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        R_FE = self._radius_free_expansion(self.sedov_taylor_begin)
        return R_FE * (t / self.sedov_taylor_begin) ** (2.0 / 5)

    def radius_inner(self, t, fraction=0.0914):
        """Inner radius  at age t  of the SNR shell.

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        return self.radius(t) * (1 - fraction)

    def luminosity_tev(self, t, energy_min="1 TeV"):
        r"""Gamma-ray luminosity above ``energy_min`` at age ``t``.

        The luminosity is assumed constant in a given age interval and zero
        before and after. The assumed spectral index is 2.1.

        The gamma-ray luminosity above 1 TeV is given by:

        .. math::
            L_{\gamma}(\geq 1TeV) \approx 10^{34} \theta
            \left(\frac{E_{SN}}{10^{51} erg}\right)
            \left(\frac{\rho_{ISM}}{1.66\cdot 10^{-24} g/cm^{3}} \right)
            \text{ s}^{-1}

        Reference: https://ui.adsabs.harvard.edu/abs/1994A%26A...287..959D (Formula (7)).

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        energy_min : `~astropy.units.Quantity`
            Lower energy limit for the luminosity.
        """
        t = Quantity(t, "yr")
        energy_min = Quantity(energy_min, "TeV")

        # Flux in 1 k distance according to Drury formula 9
        term_0 = energy_min / Quantity(1, "TeV")
        term_1 = self.e_sn / Quantity(1e51, "erg")
        term_2 = self.rho_ISM / (Quantity(1, "cm-3") * astropy.constants.m_p)
        L = self.theta * term_0 ** (1 - self.spectral_index) * term_1 * term_2

        # Corresponding luminosity
        L = np.select(
            [t <= self.sedov_taylor_begin, t <= self.sedov_taylor_end], [0, L]
        )
        return Quantity(1.0768e34, "s-1") * L

    @lazyproperty
    def sedov_taylor_begin(self):
        r"""Characteristic time scale when the Sedov-Taylor phase of the SNR's evolution begins.

        The beginning of the Sedov-Taylor phase of the SNR is defined by the condition,
        that the swept up mass of the surrounding medium equals the mass of the
        ejected mass.

        The time scale is given by:

        .. math::
            t_{begin} \approx 200
            \left(\frac{E_{SN}}{10^{51}erg}\right)^{-1/2}
            \left(\frac{M_{ej}}{M_{\odot}}\right)^{5/6}
            \left(\frac{\rho_{ISM}}{10^{-24}g/cm^3}\right)^{-1/3}
            \text{yr}
        """
        term1 = (self.e_sn / Quantity(1e51, "erg")) ** (-1.0 / 2)
        term2 = (self.m_ejecta / astropy.constants.M_sun) ** (5.0 / 6)
        term3 = (self.rho_ISM / (Quantity(1, "cm-3") * astropy.constants.m_p)) ** (
            -1.0 / 3
        )
        return Quantity(200, "yr") * term1 * term2 * term3

    @lazyproperty
    def sedov_taylor_end(self):
        r"""Characteristic time scale when the Sedov-Taylor phase of the SNR's evolution ends.

        The end of the Sedov-Taylor phase of the SNR is defined by the condition, that the
        temperature at the shock drops below T = 10^6 K.

        The time scale is given by:

        .. math::
            t_{end} \approx 43000
            \left(\frac{m}{1.66\cdot 10^{-24}g}\right)^{5/6}
            \left(\frac{E_{SN}}{10^{51}erg}\right)^{1/3}
            \left(\frac{\rho_{ISM}}{1.66\cdot 10^{-24}g/cm^3}\right)^{-1/3}
            \text{yr}
        """
        term1 = (
            3
            * astropy.constants.m_p.cgs
            / (100 * astropy.constants.k_B.cgs * self.t_stop)
        )
        term2 = (self.e_sn / self.rho_ISM) ** (2.0 / 5)
        return ((term1 * term2) ** (5.0 / 6)).to("yr")


class SNRTrueloveMcKee(SNR):
    """SNR model according to Truelove & McKee (1999).

    Reference: https://ui.adsabs.harvard.edu/abs/1999ApJS..120..299T
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Characteristic dimensions
        self.r_c = self.m_ejecta ** (1.0 / 3) * self.rho_ISM ** (-1.0 / 3)
        self.t_c = (
            self.e_sn ** (-1.0 / 2)
            * self.m_ejecta ** (5.0 / 6)
            * self.rho_ISM ** (-1.0 / 3)
        )

    def radius(self, t):
        r"""Outer shell radius at age t.

        The radius during the free expansion phase is given by:

        .. math::
            R_{SNR}(t) = 1.12R_{ch}\left(\frac{t}{t_{ch}}\right)^{2/3}

        The radius during the Sedov-Taylor phase evolves like:

        .. math::
            R_{SNR}(t) = \left[R_{SNR, ST}^{5/2} + \left(2.026\frac{E_{SN}}
            {\rho_{ISM}}\right)^{1/2}(t - t_{ST})\right]^{2/5}

        Using the characteristic dimensions:

        .. math::
            R_{ch} = M_{ej}^{1/3}\rho_{ISM}^{-1/3} \ \
            \text{and} \ \ t_{ch} = E_{SN}^{-1/2}M_{ej}^{5/6}\rho_{ISM}^{-1/3}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        t = Quantity(t, "yr")

        # Evaluate `_radius_sedov_taylor` on `t > self.sedov_taylor_begin`
        # only to avoid a warning
        r = np.empty(t.shape, dtype=np.float64)
        mask = t > self.sedov_taylor_begin
        r[mask] = self._radius_sedov_taylor(t[mask]).to_value("cm")
        r[~mask] = self._radius_free_expansion(t[~mask]).to_value("cm")
        return Quantity(r, "cm")

    def _radius_free_expansion(self, t):
        """Shock radius at age t during free expansion phase.

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        return 1.12 * self.r_c * (t / self.t_c) ** (2.0 / 3)

    def _radius_sedov_taylor(self, t):
        """Shock radius  at age t during Sedov Taylor phase.

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        term1 = self._radius_free_expansion(self.sedov_taylor_begin) ** (5.0 / 2)
        term2 = (2.026 * (self.e_sn / self.rho_ISM)) ** (1.0 / 2)
        return (term1 + term2 * (t - self.sedov_taylor_begin)) ** (2.0 / 5)

    @lazyproperty
    def sedov_taylor_begin(self):
        r"""Characteristic time scale when the Sedov-Taylor phase starts.

        Given by :math:`t_{ST} \approx 0.52 t_{ch}`.
        """
        return 0.52 * self.t_c

    def radius_reverse_shock(self, t):
        r"""Reverse shock radius at age t.

        Initially the reverse shock co-evolves with the radius of the SNR:

        .. math::
            R_{RS}(t) = \frac{1}{1.19}r_{SNR}(t)

        After a time :math:`t_{core} \simeq 0.25t_{ch}` the reverse shock reaches
        the core and then propagates as:

        .. math::
            R_{RS}(t) = \left[1.49 - 0.16 \frac{t - t_{core}}{t_{ch}} - 0.46
            \ln \left(\frac{t}{t_{core}}\right)\right]\frac{R_{ch}}{t_{ch}}t

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        t = Quantity(t, "yr")

        # Time when reverse shock reaches the "core"
        t_core = 0.25 * self.t_c

        term1 = (t - t_core) / (self.t_c)
        term2 = 1.49 - 0.16 * term1 - 0.46 * np.log(t / t_core)
        R_1 = self._radius_free_expansion(t) / 1.19
        R_RS = term2 * (self.r_c / self.t_c) * t
        r = np.where(t < t_core, R_1.to_value("cm"), R_RS.to_value("cm"))
        return Quantity(r, "cm")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from gammapy.irf.core import IRF, FoVAlignment
from gammapy.maps import MapAxis, WcsNDMap
from gammapy.irf import EDispKernelMap


class MyCustomIRF(IRF):
    tag = "myirf"
    required_axes = ["energy", "offset"]
    default_unit = u.deg


def test_irf_init_quantity():
    energy_axis = MapAxis.from_energy_bounds(10, 100, 10, unit="TeV", name="energy")
    offset_axis = MapAxis.from_bounds(0, 2.5, 5, unit="deg", name="offset")
    data = np.full((10, 5), 1)

    irf = MyCustomIRF(axes=[energy_axis, offset_axis], data=data, unit=u.deg)
    irf2 = MyCustomIRF(axes=[energy_axis, offset_axis], data=data * u.deg)

    assert np.all(irf.quantity == irf2.quantity)


def test_immutable():
    energy_axis = MapAxis.from_energy_bounds(10, 100, 10, unit="TeV", name="energy")
    offset_axis = MapAxis.from_bounds(0, 2.5, 5, unit="deg", name="offset")
    data = np.full((10, 5), 1) * u.deg
    test_irf = MyCustomIRF(
        axes=[energy_axis, offset_axis],
        data=data,
        is_pointlike=False,
        fov_alignment=FoVAlignment.RADEC,
    )

    with pytest.raises(AttributeError):
        test_irf.is_pointlike = True

    with pytest.raises(AttributeError):
        test_irf.fov_alignment = FoVAlignment.ALTAZ


def test_slice_by_idx():
    energy_axis = MapAxis.from_energy_bounds(10, 100, 10, unit="TeV", name="energy")
    offset_axis = MapAxis.from_bounds(0, 2.5, 5, unit="deg", name="offset")
    data = np.full((10, 5), 1)

    irf = MyCustomIRF(axes=[energy_axis, offset_axis], data=data, unit=u.deg)

    irf_sliced = irf.slice_by_idx({"energy": slice(3, 7)})
    assert irf_sliced.data.shape == (4, 5)
    assert irf_sliced.axes["energy"].nbin == 4

    irf_sliced = irf.slice_by_idx({"offset": slice(3, 5)})
    assert irf_sliced.data.shape == (10, 2)
    assert irf_sliced.axes["offset"].nbin == 2

    irf_sliced = irf.slice_by_idx({"energy": slice(3, 7), "offset": slice(3, 5)})
    assert irf_sliced.data.shape == (4, 2)
    assert irf_sliced.axes["offset"].nbin == 2
    assert irf_sliced.axes["energy"].nbin == 4

    with pytest.raises(ValueError) as exc_info:
        _ = irf.slice_by_idx({"energy": 3, "offset": 7})

    assert (
        str(exc_info.value)
        == "Integer indexing not supported, got {'energy': 3, 'offset': 7}"
    )


def test_cum_sum():
    energy_axis = MapAxis.from_energy_bounds(10, 100, 10, unit="TeV", name="energy")
    offset_axis = MapAxis.from_bounds(0, 2.5, 1, unit="deg", name="offset")

    data = np.full((10, 1), 1)

    irf = MyCustomIRF(axes=[energy_axis, offset_axis], data=data, unit="")
    cumsum = irf.cumsum(axis_name="offset")

    assert cumsum.unit == u.Unit("deg^2")
    assert cumsum.data[0, 0] == 2.5**2 * np.pi


def test_irfmap_downsample():
    energy_axis = MapAxis.from_energy_bounds(10, 100, 10, unit="TeV", name="energy")
    energy_true_axis = MapAxis.from_bounds(1e-1, 10, 40, unit="TeV", name="energy_true")
    m = WcsNDMap.create(npix=(200, 200), axes=[energy_axis, energy_true_axis])
    expmap = WcsNDMap.from_geom(m.geom.drop("energy"), unit="cm^2 s")
    m.data = np.random.rand(*m.data.shape)
    expmap.data = np.random.rand(*expmap.data.shape)

    weights = m.copy()
    weights.data = np.tile(
        np.array([1.0, 2.0]),
        (m.data.shape[0], m.data.shape[1], m.data.shape[2], m.data.shape[3] // 2),
    )

    irf = EDispKernelMap(m, expmap)
    irf2 = EDispKernelMap(m, None)

    # test spatial downsampling without weights
    irf3 = irf.downsample(2)

    assert irf3.edisp_map.unit == irf.edisp_map.unit
    assert irf3.edisp_map.geom.npix[0] == irf.edisp_map.geom.npix[0] / 2
    assert_allclose(
        np.mean(irf.edisp_map.data[:, :, 0:2, 0:2], axis=(2, 3)),
        irf3.edisp_map.data[:, :, 0, 0],
    )

    assert irf3.exposure_map.unit == irf.exposure_map.unit
    assert irf3.exposure_map.geom.npix[0] == irf.exposure_map.geom.npix[0] / 2
    assert_allclose(
        np.mean(irf.exposure_map.data[:, 0:2, 0:2], axis=(-1, -2)),
        irf3.exposure_map.data[:, 0, 0],
    )

    # test spatial downsampling without exposure map
    irf2.downsample(2)

    # test energy downsampling with weights
    irf5 = irf.downsample(2, axis_name="energy", weights=weights)
    assert irf5.edisp_map.data.shape[1] == irf2.edisp_map.data.shape[1] / 2
    assert_allclose(
        np.sum(
            irf2.edisp_map.data[:, 0:2, :, :] * weights.data[:, 0:2, :, :], axis=(1)
        ),
        irf5.edisp_map.data[:, 0, :, :],
    )
    assert irf5.exposure_map == irf.exposure_map

    # test spatial downsampling with weights
    irf6 = irf.downsample(2, weights=weights)
    assert_allclose(
        np.sum(
            irf.edisp_map.data[:, :, 0:2, 0:2] * weights.data[:, :, 0:2, 0:2],
            axis=(2, 3),
        )
        / np.sum(weights.data[:, :, 0:2, 0:2], axis=(2, 3)),
        irf6.edisp_map.data[:, :, 0, 0],
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Gammapy: A Python package for gamma-ray astronomy.

* Code: https://github.com/gammapy/gammapy
* Docs: https://docs.gammapy.org/

The top-level `gammapy` namespace is almost empty,
it only contains this:

::

 test                --- Run Gammapy unit tests
 __version__         --- Gammapy version string
 __maincitation__    --- Gammapy main paper
 __acknowledgment__  --- Gammapy acknowledgment


The Gammapy functionality is available for import from
the following sub-packages (e.g. `gammapy.makers`):

::

 astro        --- Astrophysical source and population models
 catalog      --- Source catalog tools
 makers       --- Data reduction functionality
 data         --- Data and observation handling
 irf          --- Instrument response functions (IRFs)
 maps         --- Sky map data structures
 modeling     --- Models and fitting
 estimators   --- High level flux estimation
 stats        --- Statistics tools
 utils        --- Utility functions and classes
"""

__all__ = ["__version__", "song", "__maincitation__", "__acknowledgment__"]

from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass
del version, PackageNotFoundError


def song(karaoke=False):
    """
    Listen to the Gammapy song.

    Make sure you listen on good headphones or speakers. You'll be not disappointed!

    Parameters
    ----------
    karaoke : bool
        Print lyrics to sing along.
    """
    import sys
    import webbrowser

    webbrowser.open("https://gammapy.org/gammapy_song.mp3")

    if karaoke:
        lyrics = (
            "\nGammapy Song Lyrics\n"
            "-------------------\n\n"
            "Gammapy, gamma-ray data analysis package\n"
            "Gammapy, prototype software CTA science tools\n\n"
            "Supernova remnants, pulsar winds, AGN, Gamma, Gamma, Gammapy\n"
            "Galactic plane survey, pevatrons, Gammapy, Gamma, Gammapy\n"
            "Gammapy, github, continuous integration, readthedocs, travis, "
            "open source project\n\n"
            "Gammapy, Gammapy\n\n"
            "Supernova remnants, pulsar winds, AGN, Gamma, Gamma, Gammapy\n"
        )

        centered = "\n".join(f"{s:^80}" for s in lyrics.split("\n"))
        sys.stdout.write(centered)


# Set the bibtex entry to the article referenced in CITATION.
def _get_bibtex():
    try:
        refs = (
            (Path(__file__).parent.parent / "CITATION")
            .read_text()
            .split("@article")[1:]
        )
        return f"@article{refs[0]}" if refs else ""
    except FileNotFoundError:
        return ""


__maincitation__ = __bibtex__ = _get_bibtex()


def _get_acknowledgment():
    try:
        text = (
            (Path(__file__).parent.parent / "CITATION")
            .read_text()
            .split(
                "If possible, we propose to add the following acknowledgment in LaTeX"
            )
        )
        ackno = text[1].split("\n\n")[1]
        return ackno
    except FileNotFoundError:
        return ""


__acknowledgment__ = _get_acknowledgment()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html

__all__ = ["Likelihood"]


# TODO: get rid of this wrapper class? Or use it in a better way?
class Likelihood:
    """Wrapper of the likelihood function used by the optimiser.

    This might become superfluous if we introduce a
    generic ``Likelihood`` interface and use that directly,
    or change the ``Fit`` class to work with ``Model``
    and ``Likelihood`` objects.

    For now, this class does the translation of parameter
    values and the parameter factors the optimiser sees.

    Parameters
    ----------
    parameters : `~gammapy.modeling.Parameters`
        Parameters with starting values.
    function : callable
        Likelihood function.
    """

    def __init__(self, function, parameters, store_trace):
        self.function = function
        self.parameters = parameters
        self.trace = []
        self.store_trace = store_trace

    def store_trace_iteration(self, total_stat):
        row = {"total_stat": total_stat}
        pars = self.parameters.free_parameters
        names = [f"par-{idx}" for idx in range(len(pars))]
        vals = dict(zip(names, pars.value))
        row.update(vals)
        self.trace.append(row)

    def fcn(self, factors):
        self.parameters.set_parameter_factors(factors)
        total_stat = self.function()

        if self.store_trace:
            self.store_trace_iteration(total_stat)

        return total_stat

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import numpy as np
from astropy.coordinates import SkyCoord
from gammapy.data import Observation
import astropy.units as u
from gammapy.maps import Map, MapAxis, WcsGeom
from gammapy.maps.utils import _check_width, _check_binsz
from gammapy.modeling.models.utils import cutout_template_models
from . import Datasets

__all__ = [
    "apply_edisp",
    "split_dataset",
    "create_map_dataset_from_dl4",
]

log = logging.getLogger(__name__)


def apply_edisp(input_map, edisp):
    """Apply energy dispersion to map. Requires "energy_true" axis.

    Parameters
    ----------
    input_map : `~gammapy.maps.Map`
        The map to be convolved with the energy dispersion.
        It must have an axis named "energy_true".
    edisp : `~gammapy.irf.EDispKernel`
        Energy dispersion matrix.

    Returns
    -------
    map : `~gammapy.maps.Map`
        Map with energy dispersion applied.

    Examples
    --------
    >>> from gammapy.irf.edisp import EDispKernel
    >>> from gammapy.datasets.utils import apply_edisp
    >>> from gammapy.maps import MapAxis, Map
    >>> import numpy as np
    >>>
    >>> axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=6, name="energy_true")
    >>> m = Map.create(
    ...     skydir=(0.8, 0.8),
    ...     width=(1, 1),
    ...     binsz=0.02,
    ...     axes=[axis],
    ...     frame="galactic"
    ... )
    >>> e_true = m.geom.axes[0]
    >>> e_reco = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3)
    >>> edisp = EDispKernel.from_diagonal_response(energy_axis_true=e_true, energy_axis=e_reco)
    >>> map_edisp = apply_edisp(m, edisp)
    >>> print(map_edisp)
    WcsNDMap
    <BLANKLINE>
        geom  : WcsGeom
        axes  : ['lon', 'lat', 'energy']
        shape : (50, 50, 3)
        ndim  : 3
        unit  :
        dtype : float64
    <BLANKLINE>
    """
    # TODO: either use sparse matrix multiplication or something like edisp.is_diagonal
    if edisp is not None:
        loc = input_map.geom.axes.index("energy_true")
        data = np.rollaxis(input_map.data, loc, len(input_map.data.shape))
        data = np.matmul(data, edisp.pdf_matrix)
        data = np.rollaxis(data, -1, loc)
        energy_axis = edisp.axes["energy"].copy(name="energy")
    else:
        data = input_map.data
        energy_axis = input_map.geom.axes["energy_true"].copy(name="energy")

    geom = input_map.geom.to_image().to_cube(axes=[energy_axis])
    return Map.from_geom(geom=geom, data=data, unit=input_map.unit)


def get_figure(fig, width, height):
    import matplotlib.pyplot as plt

    if plt.get_fignums():
        if not fig:
            fig = plt.gcf()
        fig.clf()
    else:
        fig = plt.figure(figsize=(width, height))

    return fig


def get_axes(ax1, ax2, width, height, args1, args2, kwargs1=None, kwargs2=None):
    if not ax1 and not ax2:
        kwargs1 = kwargs1 or {}
        kwargs2 = kwargs2 or {}

        fig = get_figure(None, width, height)
        ax1 = fig.add_subplot(*args1, **kwargs1)
        ax2 = fig.add_subplot(*args2, **kwargs2)
    elif not ax1 or not ax2:
        raise ValueError("Either both or no Axes must be provided")

    return ax1, ax2


def get_nearest_valid_exposure_position(exposure, position=None):
    mask_exposure = exposure > 0.0 * exposure.unit
    mask_exposure = mask_exposure.reduce_over_axes(func=np.logical_or)
    if not position:
        position = mask_exposure.geom.center_skydir
    return mask_exposure.mask_nearest_position(position)


def split_dataset(dataset, width, margin, split_template_models=True):
    """Split dataset in multiple non-overlapping analysis regions.

    Parameters
    ----------
    dataset : `~gammapy.datasets.Dataset`
        Dataset to split.
    width : `~astropy.coordinates.Angle`
        Angular size of each sub-region.
    margin : `~astropy.coordinates.Angle`
        Angular size to be added to the `width`.
        The margin should be defined such as sources outside the region of interest
        that contributes inside are well-defined.
        The mask_fit in the margin region is False and unchanged elsewhere.
    split_template_models : bool, optional
        Apply cutout to template models or not. Default is True.

    Returns
    -------
    datasets : `~gammapy.datasets.Datasets`
        Split datasets.

    Examples
    --------
    >>> from gammapy.datasets import MapDataset
    >>> from gammapy.datasets.utils import split_dataset
    >>> from gammapy.modeling.models import GaussianSpatialModel, PowerLawSpectralModel, SkyModel
    >>> import astropy.units as u
    >>> dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
    >>> # Split the dataset
    >>> width = 4 * u.deg
    >>> margin = 1 * u.deg
    >>> split_datasets = split_dataset(dataset, width, margin, split_template_models=False)
    >>> # Apply a model and split the dataset
    >>> spatial_model = GaussianSpatialModel()
    >>> spectral_model = PowerLawSpectralModel()
    >>> sky_model = SkyModel(
    ...     spatial_model=spatial_model, spectral_model=spectral_model, name="test-model"
    ... )
    >>> dataset.models = [sky_model]
    >>> split_datasets = split_dataset(
    ...     dataset, width=width, margin=margin, split_template_models=True
    ... )
    """
    if margin >= width / 2.0:
        raise ValueError("margin should be lower than width/2.")

    geom = dataset.counts.geom.to_image()
    pixel_width = np.ceil((width / geom.pixel_scales).to_value("")).astype(int)

    ilon = range(0, geom.data_shape[1], pixel_width[1])
    ilat = range(0, geom.data_shape[0], pixel_width[0])

    datasets = Datasets()
    for il in ilon:
        for ib in ilat:
            l, b = geom.pix_to_coord((il, ib))
            cutout_kwargs = dict(
                position=SkyCoord(l, b, frame=geom.frame), width=width + 2 * margin
            )
            d = dataset.cutout(**cutout_kwargs)

            # apply mask
            geom_cut = d.counts.geom
            geom_cut_image = geom_cut.to_image()
            ilgrid, ibgrid = np.meshgrid(
                range(geom_cut_image.data_shape[1]), range(geom_cut_image.data_shape[0])
            )
            il_cut, ib_cut = geom_cut_image.coord_to_pix((l, b))
            mask = (ilgrid >= il_cut - pixel_width[1] / 2.0) & (
                ibgrid >= ib_cut - pixel_width[0] / 2.0
            )
            if il == ilon[-1]:
                mask &= ilgrid <= il_cut + pixel_width[1] / 2.0
            else:
                mask &= ilgrid < il_cut + pixel_width[1] / 2.0

            if ib == ilat[-1]:
                mask &= ibgrid <= ib_cut + pixel_width[0] / 2.0
            else:
                mask &= ibgrid < ib_cut + pixel_width[0] / 2.0
            mask = np.expand_dims(mask, 0)
            mask = np.repeat(mask, geom_cut.data_shape[0], axis=0)
            d.mask_fit = Map.from_geom(geom_cut, data=mask)
            if dataset.mask_fit is not None:
                d.mask_fit &= dataset.mask_fit.interp_to_geom(
                    geom_cut, method="nearest"
                )

            # template models cutout (should limit memory usage in parallel)
            if split_template_models:
                d.models = cutout_template_models(
                    dataset.models, cutout_kwargs, [d.name]
                )
            else:
                d.models = dataset.models
            datasets.append(d)
    return datasets


def create_map_dataset_from_dl4(data, geom=None, energy_axis_true=None, name=None):
    """Create a map dataset from a map dataset or an observation containing DL4 IRFs.

    Parameters
    ----------
    data : `~gammapy.dataset.MapDataset` or `~gammapy.data.Observation`
        MapDataset or Observation containing DL4 IRFs
    geom : `~gammapy.maps.WcsGeom`, optional
        Output dataset maps geometry. The default is None, and it is derived from IRFs
    energy_axis_true : `~gammapy.maps.MapAxis`, optional
        True energy axis used for IRF maps. The default is None, and it is derived from IRFs
    name : str, optional
        Dataset name. The default is None, and the name is randomly generated.

    Returns
    -------
    dataset : `~gammapy.datasets.MapDataset`
        Map dataset.
    """
    from gammapy.makers import MapDatasetMaker
    from gammapy.datasets import MapDataset

    # define target geom
    if geom is None:
        if isinstance(data, Observation):
            geom_image = data.aeff.geom.to_image()
        elif isinstance(data, MapDataset):
            geom_image = data.exposure.geom.to_image()

        geom = geom_image.to_cube([data.edisp.edisp_map.geom.axes["energy"]])

    energy_axis = geom.axes["energy"]

    if energy_axis_true is None:
        energy_axis_true = data.edisp.edisp_map.geom.axes["energy_true"]

    # ensure that DL4 IRFs have the axes
    rad_axis = data.psf.psf_map.geom.axes["rad"]
    if data.psf.energy_name == "energy":
        geom_psf = data.psf.psf_map.geom.to_image().to_cube([rad_axis, energy_axis])
    else:
        geom_psf = data.psf.psf_map.geom.to_image().to_cube(
            [rad_axis, energy_axis_true]
        )

    geom_edisp = data.edisp.edisp_map.geom.to_image().to_cube(
        [energy_axis, energy_axis_true]
    )
    geom_exposure = geom.to_image().to_cube([energy_axis_true])

    # create dataset and run data reduction / irfs interpolation
    dataset = MapDataset.from_geoms(
        geom,
        geom_exposure=geom_exposure,
        geom_psf=geom_psf,
        geom_edisp=geom_edisp,
        name=name,
    )

    selection = ["exposure", "edisp", "psf"]
    if isinstance(data, Observation) and data.events:
        selection.append("counts")
    if isinstance(data, Observation) and data.bkg:
        selection.append("background")

    maker = MapDatasetMaker(selection=selection)
    dataset = maker.run(dataset, data)

    if isinstance(data, MapDataset) and data.counts:
        if dataset.counts.geom == data.counts.geom:
            dataset.counts.data = data.counts.data
        else:
            raise ValueError(
                "Counts geom of input MapDataset and target geom must be identical"
            )

    if not dataset.background:
        dataset.background = Map.from_geom(geom, data=0.0)

    if dataset.edisp.exposure_map and np.all(dataset.edisp.exposure_map.data) == 0.0:
        dataset.edisp.exposure_map.quantity = dataset.psf.exposure_map.quantity

    return dataset


def create_global_dataset(
    datasets,
    name=None,
    position=None,
    binsz=None,
    width=None,
    energy_min=None,
    energy_max=None,
    energy_true_min=None,
    energy_true_max=None,
    nbin_per_decade=None,
):
    """Create an empty dataset encompassing the input datasets.

    Parameters
    ----------
    datasets : `~gammapy.datasets.Datasets`
        Datasets
    name : str, optional
        Name of the output dataset. Default is None.
    position : `~astropy.coordinates.SkyCoord`
        Center position of the output dataset.
        Default is None, and the average position is taken.
    binsz : float or tuple or list, optional
        Map pixel size in degrees.
        Default is None, the minimum bin size is taken.
    width : float or tuple or list or string, optional
        Width of the map in degrees.
        Default is None, in that case it is derived as
        the maximal width + twice the maximal separation between datasets and position.
    energy_min :  `~astropy.units.Quantity`
        Energy range.
        Default is None, the minimum energy is taken.
    energy_max :  `~astropy.units.Quantity`
        Energy range.
        Default is None, the maximum energy is taken.
    energy_true_min, energy_true_max :  `~astropy.units.Quantity`,  `~astropy.units.Quantity`
        True energy range. If None, minimum and maximum energies of all geometries is used.
        Default is None.
    nbin_per_decade : int
        number of energy bins per decade.
        Default is None, the maximum is taken.

    Returns
    -------
    datasets : `~gammapy.datasets.MapDatset`
        Empy global dataset.
    """
    from gammapy.datasets import MapDataset

    if position is None:
        frame = datasets[0].counts.geom.frame
        positions = SkyCoord(
            [d.counts.geom.center_skydir.transform_to(frame) for d in datasets]
        )
        position = SkyCoord(positions.cartesian.mean(), frame="icrs")
        position = SkyCoord(position.ra, position.dec, frame="icrs").transform_to(
            frame
        )  # drop fake distance

    binsz_list = []
    width_list = []
    energy_min_list = []
    energy_max_list = []
    energy_true_min_list = []
    energy_true_max_list = []
    nbin_per_decade_list = []
    for d in datasets:
        binsz_list.append(np.abs(d.counts.geom.pixel_scales).min())
        width_list.append(
            d.counts.geom.width.max()
            + 2 * d.counts.geom.center_skydir.separation(position)
        )
        energy_min_list.append(d.counts.geom.axes["energy"].edges.min())
        energy_max_list.append(d.counts.geom.axes["energy"].edges.max())
        energy_true_min_list.append(d.exposure.geom.axes["energy_true"].edges.min())
        energy_true_max_list.append(d.exposure.geom.axes["energy_true"].edges.max())
        ndecade = np.log10(energy_true_max_list[-1].value) - np.log10(
            energy_true_min_list[-1].value
        )
        nbins = len(d.exposure.geom.axes[0].center)
        nbin_per_decade_list.append(np.ceil(nbins / ndecade))

    if binsz is None:
        binsz = np.min(u.Quantity(binsz_list))
    binsz = _check_binsz(binsz)
    if width is None:
        width = np.max(u.Quantity(width_list))
    width = _check_width(width)
    energy_true_min = (
        energy_true_min
        if energy_true_min is not None
        else u.Quantity(energy_true_min_list).min()
    )
    if energy_true_max is None:
        energy_true_max = np.max(u.Quantity(energy_true_max_list))
    if energy_min is None:
        energy_min = np.min(u.Quantity(energy_min_list))
    if energy_max is None:
        energy_max = np.max(u.Quantity(energy_max_list))

    nbin_per_decade = (
        nbin_per_decade if nbin_per_decade is not None else np.max(nbin_per_decade_list)
    )
    energy_axis = MapAxis.from_energy_bounds(
        energy_min, energy_max, nbin_per_decade, unit="TeV", per_decade=True
    )

    geom = WcsGeom.create(
        skydir=position,
        binsz=binsz,
        width=width,
        frame=position.frame,
        proj=datasets[0].counts.geom.projection,
        axes=[energy_axis],
    )

    energy_axis_true = MapAxis.from_energy_bounds(
        energy_true_min,
        energy_true_max,
        nbin_per_decade,
        unit="TeV",
        name="energy_true",
        per_decade=True,
    )
    return MapDataset.create(geom=geom, energy_axis_true=energy_axis_true, name=name)


class set_and_restore_mask_fit:
    """Context manager to set a `mask_fit` on dataset and restore the initial mask.

    Parameters
    ----------
    datasets : `~gammapy.datasets.datasets`
        the Datasets to apply the energy mask to.
    energy_min : `~astropy.units.Quantity`, optional
        minimum energy.
    energy_min : `~astropy.units.Quantity`, optional
        maximum energy.
    round_to_edge: bool, optional
        Whether to round `energy_min` and `energy_max` to the closest axis bin value.
        See `~gammapy.maps.MapAxis.round`. Default is False.
    """

    def __init__(
        self,
        datasets,
        mask_fit=None,
        energy_min=None,
        energy_max=None,
        round_to_edge=False,
    ):
        self.energy_min = energy_min
        self.energy_max = energy_max
        self.round_to_edge = round_to_edge
        self.mask_fit = mask_fit
        self.datasets = datasets
        self.mask_fits = [dataset.mask_fit for dataset in datasets]

    def __enter__(self):
        datasets = Datasets()
        for dataset in self.datasets:
            mask_fit = dataset._geom.energy_mask(
                self.energy_min, self.energy_max, self.round_to_edge
            )
            if self.mask_fit is not None:
                mask_fit *= self.mask_fit.interp_to_geom(
                    mask_fit.geom, method="nearest"
                )
            dataset.mask_fit = mask_fit
            if np.any(mask_fit.data):
                datasets.append(dataset)
            else:
                log.info(
                    f"Dataset {dataset.name} does not contribute in the energy range"
                )
        return datasets

    def __exit__(self, type, value, traceback):
        for dataset, mask_fit in zip(self.datasets, self.mask_fits):
            dataset.mask_fit = mask_fit



r"""
.. _naima-spectral-model:

Naima spectral model
====================

This class provides an interface with the models defined in the naima models module.

The model accepts as a positional argument a `Naima <https://naima.readthedocs.io/en/latest/>`_
radiative `~naima.models` instance, used to compute the non-thermal emission from populations of
relativistic electrons or protons due to interactions with the ISM or with radiation and magnetic fields.

One of the advantages provided by this class consists in the possibility of performing a maximum
likelihood spectral fit of the model's parameters directly on observations, as opposed to the MCMC
`fit to flux points <https://naima.readthedocs.io/en/latest/mcmc.html>`_ featured in
Naima. All the parameters defining the parent population of charged particles are stored as
`~gammapy.modeling.Parameter` and left free by default. In case that the radiative model is
`~naima.radiative.Synchrotron`, the magnetic field strength may also be fitted. Parameters can be
freezed/unfreezed before the fit, and maximum/minimum values can be set to limit the parameters space to
the physically interesting region.


"""

# %%
# Example plot
# ------------
# Here we create and plot a spectral model that convolves an `~gammapy.modeling.models.ExpCutoffPowerLawSpectralModel`
# electron distribution with an `InverseCompton` radiative model, in the presence of multiple seed photon fields.

from astropy import units as u
import matplotlib.pyplot as plt
import naima
from gammapy.modeling.models import Models, NaimaSpectralModel, SkyModel

particle_distribution = naima.models.ExponentialCutoffPowerLaw(
    1e30 / u.eV, 10 * u.TeV, 3.0, 30 * u.TeV
)
radiative_model = naima.radiative.InverseCompton(
    particle_distribution,
    seed_photon_fields=["CMB", ["FIR", 26.5 * u.K, 0.415 * u.eV / u.cm**3]],
    Eemin=100 * u.GeV,
)

model = NaimaSpectralModel(radiative_model, distance=1.5 * u.kpc)

opts = {
    "energy_bounds": [10 * u.GeV, 80 * u.TeV],
    "sed_type": "e2dnde",
}

# Plot the total inverse Compton emission
model.plot(label="IC (total)", **opts)

# Plot the separate contributions from each seed photon field
for seed, ls in zip(["CMB", "FIR"], ["-", "--"]):
    model = NaimaSpectralModel(radiative_model, seed=seed, distance=1.5 * u.kpc)
    model.plot(label=f"IC ({seed})", ls=ls, color="gray", **opts)

plt.legend(loc="best")
plt.grid(which="both")

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(spectral_model=model, name="naima-model")
models = Models([model])

print(models.to_yaml())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from scipy.stats import norm
import astropy.units as u
from astropy.table import Table
import matplotlib.pyplot as plt
from gammapy.maps import Map, MapAxis, WcsNDMap
from gammapy.utils.random import get_random_state
from gammapy.utils.testing import mpl_plot_check, requires_data
from gammapy.visualization import (
    add_colorbar,
    plot_contour_line,
    plot_distribution,
    plot_map_rgb,
    plot_theta_squared_table,
)


@requires_data()
def test_add_colorbar():
    map_ = Map.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")

    fig, ax = plt.subplots()
    with mpl_plot_check():
        img = ax.imshow(map_.sum_over_axes().data[0, :, :])
        add_colorbar(img, ax=ax, label="Colorbar label")

    axes_loc = {"position": "left", "size": "2%", "pad": "15%"}
    fig, ax = plt.subplots()
    with mpl_plot_check():
        img = ax.imshow(map_.sum_over_axes().data[0, :, :])
        add_colorbar(img, ax=ax, axes_loc=axes_loc)

    kwargs = {"use_gridspec": False, "orientation": "horizontal"}
    fig, ax = plt.subplots()
    with mpl_plot_check():
        img = ax.imshow(map_.sum_over_axes().data[0, :, :])
        cbar = add_colorbar(img, ax=ax, **kwargs)
        assert cbar.orientation == "horizontal"


def test_map_panel_plotter():
    t = np.linspace(0.0, 6.1, 10)
    x = np.cos(t)
    y = np.sin(t)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    with mpl_plot_check():
        plot_contour_line(ax, x, y)

    x = np.append(x, x[0])
    y = np.append(y, y[0])
    with mpl_plot_check():
        plot_contour_line(ax, x, y)


def test_plot_theta2_distribution():
    table = Table()
    table["theta2_min"] = [0, 0.1]
    table["theta2_max"] = [0.1, 0.2]

    for column in [
        "counts",
        "counts_off",
        "excess",
        "excess_errp",
        "excess_errn",
        "sqrt_ts",
    ]:
        table[column] = [1, 1]

    # open a new figure to avoid
    plt.figure()
    plot_theta_squared_table(table=table)


@requires_data()
def test_plot_map_rgb():
    map_ = Map.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")

    with pytest.raises(ValueError):
        plot_map_rgb(map_)

    with pytest.raises(ValueError):
        plot_map_rgb(map_.sum_over_axes(keepdims=False))

    axis = MapAxis([0, 1, 2, 3], node_type="edges")
    map_allsky = WcsNDMap.create(binsz=10 * u.deg, axes=[axis])

    # Astropy 7 does not support uniformly 0 maps
    map_allsky.data += 1

    with mpl_plot_check():
        plot_map_rgb(map_allsky)

    axis_rgb = MapAxis.from_energy_edges(
        [0.1, 0.2, 0.5, 10], unit=u.TeV, name="energy", interp="log"
    )
    map_ = map_.resample_axis(axis_rgb)
    kwargs = {"stretch": 0.5, "Q": 1, "minimum": 0.15}
    with mpl_plot_check():
        plot_map_rgb(map_, **kwargs)


def test_plot_distribution():
    random_state = get_random_state(0)
    array = random_state.normal(0, 1, 10000)

    array_2d = array.reshape(1, 100, 100)

    energy_axis = MapAxis.from_energy_edges([1, 10] * u.TeV)

    map_ = WcsNDMap.create(npix=(100, 100), axes=[energy_axis])
    map_.data = array_2d
    mask_ = WcsNDMap.create(npix=(100, 100), dtype="bool")
    mask_.data[:50, :] = True

    energy_axis_10 = MapAxis.from_energy_bounds(1 * u.TeV, 10 * u.TeV, 10)
    map_empty = WcsNDMap.create(npix=(100, 100), axes=[energy_axis_10])

    def fit_func(x, mu, sigma):
        return norm.pdf(x, mu, sigma)

    with mpl_plot_check():
        axes, res = plot_distribution(
            wcs_map=map_, mask=mask_, func=fit_func, kwargs_hist={"bins": 40}
        )

        assert axes.shape == (1,)
        assert "info_dict" in res[0]
        assert ["fvec", "nfev", "fjac", "ipvt", "qtf"] == list(
            (res[0].get("info_dict").keys())
        )
        assert res[0].get("param") is not None
        assert res[0].get("covar") is not None

        axes, res = plot_distribution(map_empty)

        assert res == []
        assert axes.shape == (4, 3)

        axes, res = plot_distribution(
            wcs_map=map_, mask=mask_, func="norm", kwargs_hist={"bins": 40}
        )

        fig, ax = plt.subplots(nrows=1, ncols=1)
        plot_distribution(map_empty, ax=ax)

        # test mask exceptions
        mask_2 = WcsNDMap.create(npix=(200, 200), dtype="bool")
        mask_2.data[:50, :] = True
        with pytest.raises(ValueError):
            axes, res = plot_distribution(
                wcs_map=map_, mask=mask_2, func="norm", kwargs_hist={"bins": 40}
            )

        mask_3 = WcsNDMap.create(npix=(100, 100), dtype="float")
        mask_3.data = np.ones_like(mask_3.data)
        with pytest.raises(ValueError):
            axes, res = plot_distribution(
                wcs_map=map_, mask=mask_3, func="norm", kwargs_hist={"bins": 40}
            )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from copy import deepcopy
import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.utils import classproperty
from gammapy.data import GTI
from gammapy.maps import Map, Maps, TimeMapAxis
from gammapy.modeling.models import (
    Models,
    PowerLawSpectralModel,
    SkyModel,
    SpectralModel,
)
from gammapy.utils.scripts import make_path

__all__ = ["FluxMaps"]

log = logging.getLogger(__name__)


DEFAULT_UNIT = {
    "dnde": u.Unit("cm-2 s-1 TeV-1"),
    "e2dnde": u.Unit("erg cm-2 s-1"),
    "flux": u.Unit("cm-2 s-1"),
    "eflux": u.Unit("erg cm-2 s-1"),
    "norm": u.Unit(""),
}

REQUIRED_MAPS = {
    "dnde": ["dnde"],
    "e2dnde": ["e2dnde"],
    "flux": ["flux"],
    "eflux": ["eflux"],
    "likelihood": ["norm"],
}

REQUIRED_COLUMNS = {
    "dnde": ["e_ref", "dnde"],
    "e2dnde": ["e_ref", "e2dnde"],
    "flux": ["e_min", "e_max", "flux"],
    "eflux": ["e_min", "e_max", "eflux"],
    "likelihood": [
        "e_min",
        "e_max",
        "e_ref",
        "ref_dnde",
        "ref_flux",
        "ref_eflux",
        "norm",
    ],
}


REQUIRED_QUANTITIES_SCAN = ["stat_scan", "stat"]

OPTIONAL_QUANTITIES = {
    "dnde": ["dnde_err", "dnde_errp", "dnde_errn", "dnde_ul"],
    "e2dnde": ["e2dnde_err", "e2dnde_errp", "e2dnde_errn", "e2dnde_ul"],
    "flux": ["flux_err", "flux_errp", "flux_errn", "flux_ul", "flux_sensitivity"],
    "eflux": ["eflux_err", "eflux_errp", "eflux_errn", "eflux_ul"],
    "likelihood": ["norm_err", "norm_errn", "norm_errp", "norm_ul", "norm_sensitivity"],
}

VALID_QUANTITIES = [
    "norm",
    "norm_err",
    "norm_errn",
    "norm_errp",
    "norm_ul",
    "norm_sensitivity",
    "ts",
    "sqrt_ts",
    "npred",
    "npred_excess",
    "stat",
    "stat_null",
    "stat_scan",
    "dnde_scan_values",
    "niter",
    "is_ul",
    "counts",
    "success",
    "n_dof",
    "alpha",
    "acceptance_on",
    "acceptance_off",
]


OPTIONAL_QUANTITIES_COMMON = [
    "ts",
    "sqrt_ts",
    "npred",
    "npred_excess",
    "stat",
    "stat_null",
    "stat_scan",
    "dnde_scan_values",
    "niter",
    "is_ul",
    "counts",
    "success",
    "n_dof",
]


class FluxMaps:
    """A flux map / points container.

    It contains a set of `~gammapy.maps.Map` objects that store the estimated
    flux as a function of energy as well as associated quantities (typically
    errors, upper limits, delta TS and possibly raw quantities such counts,
    excesses etc.). It also contains a reference model to convert the flux
    values in different formats. Usually, this should be the model used to
    produce the flux map.

    The associated map geometry can use a `~gammapy.maps.RegionGeom` to store the equivalent
    of flux points, or a `~gammapy.maps.WcsGeom`/`~gammapy.maps.HpxGeom` to store an energy dependent flux map.

    The container relies internally on the 'Likelihood' SED type defined in
    :ref:`gadf:flux-points` and offers convenience properties to convert to
    other flux formats, namely: ``dnde``, ``flux``, ``eflux`` or ``e2dnde``.
    The conversion is done according to the reference model spectral shape.

    Parameters
    ----------
    data : dict of `~gammapy.maps.Map`
        The maps dictionary. Expected entries are the following:

        * norm : the norm factor.
        * norm_err : optional, the error on the norm factor.
        * norm_errn : optional, the negative error on the norm factor.
        * norm_errp : optional, the positive error on the norm factor.
        * norm_ul : optional, the upper limit on the norm factor.
        * norm_scan : optional, the norm values of the test statistic scan.
        * stat_scan : optional, the test statistic scan values.
        * ts : optional, the delta test statistic associated with the flux value.
        * sqrt_ts : optional, the square root of the test statistic, when relevant.
        * success : optional, a boolean tagging the validity of the estimation.
        * n_dof : optional, the number of degrees of freedom used in TS computation
        * alpha : optional, normalisation factor to accounts for differences between the test region and the background
        * acceptance_off : optional, acceptance from the off region
        * acceptance_on : optional, acceptance from the on region

    reference_model : `~gammapy.modeling.models.SkyModel`, optional
        The reference model to use for conversions. If None, a model consisting
        of a point source with a power law spectrum of index 2 is assumed.
    meta : dict, optional
        Dict of metadata.
    gti : `~gammapy.data.GTI`, optional
        Maps GTI information.
    filter_success_nan : boolean, optional
        Set fitted norm and error to NaN when the fit has not succeeded.
    """

    _expand_slice = (slice(None), np.newaxis, np.newaxis)

    def __init__(
        self, data, reference_model, meta=None, gti=None, filter_success_nan=True
    ):
        self._data = data

        if isinstance(reference_model, SpectralModel):
            reference_model = SkyModel(reference_model)

        self._reference_model = reference_model

        if meta is None:
            meta = {}

        meta.setdefault("sed_type_init", "likelihood")
        self.meta = meta
        self.gti = gti
        self._filter_success_nan = filter_success_nan

    @property
    def filter_success_nan(self):
        return self._filter_success_nan

    @filter_success_nan.setter
    def filter_success_nan(self, value):
        self._filter_success_nan = value

    @property
    def available_quantities(self):
        """Available quantities."""
        return list(self._data.keys())

    @staticmethod
    def all_quantities(sed_type):
        """All quantities allowed for a given SED type.

        Parameters
        ----------
        sed_type : {"likelihood", "dnde", "e2dnde", "flux", "eflux"}
            SED type.

        Returns
        -------
        list : list of str
            All allowed quantities for a given SED type.
        """
        quantities = []
        quantities += REQUIRED_MAPS[sed_type]
        quantities += OPTIONAL_QUANTITIES[sed_type]
        quantities += OPTIONAL_QUANTITIES_COMMON

        if sed_type == "likelihood":
            quantities += REQUIRED_QUANTITIES_SCAN

        return quantities

    @staticmethod
    def _validate_data(data, sed_type, check_scan=False):
        """Check that map input is valid and correspond to one of the SED type."""
        try:
            keys = data.keys()
            required = set(REQUIRED_MAPS[sed_type])
        except KeyError:
            raise ValueError(f"Unknown SED type: '{sed_type}'")

        if check_scan:
            required = required.union(REQUIRED_QUANTITIES_SCAN)

        if not required.issubset(keys):
            missing = required.difference(keys)
            raise ValueError(
                "Missing data / column for SED type '{}':" " {}".format(
                    sed_type, missing
                )
            )

    # TODO: add support for scan
    def _check_quantity(self, quantity):
        if quantity not in self.available_quantities:
            raise AttributeError(
                f"Quantity '{quantity}' is not defined on current flux estimate."
            )

    @staticmethod
    def _guess_sed_type(quantities):
        """Guess SED type from table content."""
        valid_sed_types = list(REQUIRED_COLUMNS.keys())
        for sed_type in valid_sed_types:
            required = set(REQUIRED_COLUMNS[sed_type])
            if required.issubset(quantities):
                return sed_type

    @property
    def is_convertible_to_flux_sed_type(self):
        """Check whether differential SED type is convertible to integral SED type."""
        if self.sed_type_init in ["dnde", "e2dnde"]:
            return self.energy_axis.node_type == "edges"

        return True

    @property
    def has_ul(self):
        """Whether the flux estimate has norm_ul defined."""
        return "norm_ul" in self._data

    @property
    def has_any_ts(self):
        """Whether the flux estimate has either sqrt(TS) or test statistic defined."""
        return any(_ in self._data for _ in ["ts", "sqrt_ts"])

    @property
    def has_stat_profiles(self):
        """Whether the flux estimate has test statistic profiles."""
        return "stat_scan" in self._data

    @property
    def has_success(self):
        """Whether the flux estimate has the fit status."""
        return "success" in self._data

    @property
    def n_sigma(self):
        """Number of sigma."""
        return self.meta.get("n_sigma", 1)

    @property
    def n_sigma_ul(self):
        """Number of sigma UL."""
        return self.meta.get("n_sigma_ul")

    @property
    def sqrt_ts_threshold_ul(self):
        """sqrt(TS) threshold for upper limits."""
        return self.meta.get("sqrt_ts_threshold_ul", 2)

    @sqrt_ts_threshold_ul.setter
    def sqrt_ts_threshold_ul(self, value):
        """sqrt(TS) threshold for upper limits.

        Parameters
        ----------
        value : int
            Threshold value in sqrt(TS) for upper limits.
        """
        self.meta["sqrt_ts_threshold_ul"] = value

        if self.has_any_ts:
            self.is_ul = self.sqrt_ts < self.sqrt_ts_threshold_ul
        else:
            raise ValueError("Either TS or sqrt_ts is required to set the threshold")

    @property
    def sed_type_init(self):
        """Initial SED type."""
        return self.meta.get("sed_type_init")

    @property
    def sed_type_plot_default(self):
        """Initial SED type."""
        if self.sed_type_init == "likelihood":
            return "dnde"

        return self.sed_type_init

    @property
    def geom(self):
        """Reference map geometry as a `~gammapy.maps.Geom`."""
        return self.norm.geom

    @property
    def energy_axis(self):
        """Energy axis as a `~gammapy.maps.MapAxis`."""
        return self.geom.axes["energy"]

    @classproperty
    def reference_model_default(self):
        """Reference model default as a `~gammapy.modeling.models.SkyModel`."""
        return SkyModel(PowerLawSpectralModel(index=2))

    @property
    def reference_model(self):
        """Reference model as a `~gammapy.modeling.models.SkyModel`."""
        return self._reference_model

    @property
    def reference_spectral_model(self):
        """Reference spectral model as a `SpectralModel`."""
        return self.reference_model.spectral_model

    @property
    def energy_ref(self):
        """Reference energy.

        Defined by `energy_ref` column in `FluxPoints.table` or computed as log
        center, if `energy_min` and `energy_max` columns are present in `FluxEstimate.data`.

        Returns
        -------
        energy_ref : `~astropy.units.Quantity`
            Reference energy.
        """
        return self.energy_axis.center

    @property
    def energy_min(self):
        """Energy minimum.

        Returns
        -------
        energy_min : `~astropy.units.Quantity`
            Lower bound of energy bin.
        """
        return self.energy_axis.edges[:-1]

    @property
    def energy_max(self):
        """Energy maximum.

        Returns
        -------
        energy_max : `~astropy.units.Quantity`
            Upper bound of energy bin.
        """
        return self.energy_axis.edges[1:]

    # TODO: keep or remove?
    @property
    def niter(self):
        """Number of iterations of fit."""
        self._check_quantity("niter")
        return self._data["niter"]

    @property
    def success(self):
        """Fit success flag."""
        self._check_quantity("success")
        return self._data["success"]

    @property
    def is_ul(self):
        """Whether data is an upper limit."""
        # TODO: make this a well defined behaviour
        is_ul = self.norm.copy(data=False)

        if "is_ul" in self._data:
            is_ul = self._data["is_ul"]
        elif self.has_any_ts and self.has_ul:
            is_ul.data = self.sqrt_ts.data < self.sqrt_ts_threshold_ul
        elif self.has_ul:
            is_ul.data = np.isfinite(self.norm_ul)
        else:
            is_ul.data = np.isnan(self.norm)

        return is_ul

    @is_ul.setter
    def is_ul(self, value):
        """Whether data is an upper limit.

        Parameters
        ----------
        value : `~Map`
            Boolean map.
        """
        if not isinstance(value, Map):
            value = self.norm.copy(data=value)

        self._data["is_ul"] = value

    @property
    def counts(self):
        """Predicted counts null hypothesis."""
        self._check_quantity("counts")
        return self._data["counts"]

    @property
    def npred(self):
        """Predicted counts from best fit hypothesis."""
        self._check_quantity("npred")
        return self._data["npred"]

    @property
    def npred_background(self):
        """Predicted background counts from best fit hypothesis."""
        self._check_quantity("npred")
        self._check_quantity("npred_excess")
        return self.npred - self.npred_excess

    @property
    def npred_excess(self):
        """Predicted excess count from best fit hypothesis."""
        self._check_quantity("npred_excess")
        return self._data["npred_excess"]

    def _expand_dims(self, data):
        # TODO: instead make map support broadcasting
        axes = self.npred.geom.axes
        # here we need to rely on broadcasting
        if "dataset" in axes.names:
            idx = axes.index_data("dataset")
            data = np.expand_dims(data, axis=idx)
        return data

    @staticmethod
    def _use_center_as_labels(input_map):
        """Change the node_type of the input map."""
        energy_axis = input_map.geom.axes["energy"]
        energy_axis.use_center_as_plot_labels = True
        return input_map

    @property
    def npred_excess_ref(self):
        """Predicted excess reference counts."""
        return self.npred_excess / self._expand_dims(self.norm.data)

    @property
    def npred_excess_err(self):
        """Predicted excess counts error."""
        return self.npred_excess_ref * self._expand_dims(self.norm_err.data)

    @property
    def npred_excess_errp(self):
        """Predicted excess counts positive error."""
        return self.npred_excess_ref * self._expand_dims(self.norm_errp.data)

    @property
    def npred_excess_errn(self):
        """Predicted excess counts negative error."""
        return self.npred_excess_ref * self._expand_dims(self.norm_errn.data)

    @property
    def npred_excess_ul(self):
        """Predicted excess counts upper limits."""
        return self.npred_excess_ref * self._expand_dims(self.norm_ul.data)

    @property
    def stat_scan(self):
        """Fit statistic scan value."""
        self._check_quantity("stat_scan")
        return self._data["stat_scan"]

    @property
    def dnde_scan_values(self):
        """Fit statistic norm scan values."""
        self._check_quantity("dnde_scan_values")
        return self._data["dnde_scan_values"]

    @property
    def stat(self):
        """Fit statistic value."""
        self._check_quantity("stat")
        return self._data["stat"]

    @property
    def stat_null(self):
        """Fit statistic value for the null hypothesis."""
        self._check_quantity("stat_null")
        return self._data["stat_null"]

    @property
    def ts(self):
        """Test statistic map as a `~gammapy.maps.Map` object."""
        self._check_quantity("ts")
        return self._data["ts"]

    @property
    def ts_scan(self):
        """Test statistic scan as a `~gammapy.maps.Map` object."""
        return self.stat_scan - np.expand_dims(self.stat.data, 2)

    # TODO: always derive sqrt(TS) from TS?
    @property
    def sqrt_ts(self):
        r"""sqrt(TS) as defined by:

        .. math::

            \sqrt{TS} = \left \{
            \begin{array}{ll}
              -\sqrt{TS} & : \text{if} \ norm < 0 \\
              \sqrt{TS} & : \text{else}
            \end{array}
            \right.

        Returns
        -------
        sqrt_ts : `Map`
            sqrt(TS) map.
        """
        if "sqrt_ts" in self._data:
            return self._data["sqrt_ts"]
        else:
            with np.errstate(invalid="ignore", divide="ignore"):
                ts = np.clip(self.ts.data, 0, None)
                data = np.where(self.norm > 0, np.sqrt(ts), -np.sqrt(ts))
                return Map.from_geom(geom=self.geom, data=data)

    @property
    def norm(self):
        """Norm values."""
        return self._filter_convergence_failure(self._data["norm"])

    @property
    def norm_err(self):
        """Norm error."""
        self._check_quantity("norm_err")
        return self._filter_convergence_failure(self._data["norm_err"])

    @property
    def norm_errn(self):
        """Negative norm error."""
        self._check_quantity("norm_errn")
        return self._data["norm_errn"]

    @property
    def norm_errp(self):
        """Positive norm error."""
        self._check_quantity("norm_errp")
        return self._data["norm_errp"]

    @property
    def norm_ul(self):
        """Norm upper limit."""
        self._check_quantity("norm_ul")
        return self._data["norm_ul"]

    @property
    def norm_sensitivity(self):
        """Norm sensitivity."""
        self._check_quantity("norm_sensitivity")
        return self._data["norm_sensitivity"]

    @property
    def n_dof(self):
        """Number of degrees of freedom of the fit per energy bin."""
        self._check_quantity("n_dof")
        return self._data["n_dof"]

    @property
    def alpha(self):
        """The normalisation, alpha, for differences between the on and off regions."""
        self._check_quantity("alpha")
        return self._data["alpha"]

    @property
    def acceptance_on(self):
        """The acceptance in the on region."""
        self._check_quantity("acceptance_on")
        return self._data["acceptance_on"]

    @property
    def acceptance_off(self):
        """The acceptance in the off region."""
        self._check_quantity("acceptance_off")
        return self._data["acceptance_off"]

    @property
    def dnde_ref(self):
        """Reference differential flux."""
        result = self.reference_spectral_model(self.energy_axis.center)
        return result[self._expand_slice]

    @property
    def e2dnde_ref(self):
        """Reference differential flux * energy ** 2."""
        energy = self.energy_axis.center
        result = self.reference_spectral_model(energy) * energy**2
        return result[self._expand_slice]

    @property
    def flux_ref(self):
        """Reference integral flux."""
        if not self.is_convertible_to_flux_sed_type:
            raise ValueError(
                "Missing energy range definition, cannot convert to SED type 'flux'."
            )

        energy_min = self.energy_axis.edges[:-1]
        energy_max = self.energy_axis.edges[1:]
        result = self.reference_spectral_model.integral(energy_min, energy_max)
        return result[self._expand_slice]

    @property
    def eflux_ref(self):
        """Reference energy flux."""
        if not self.is_convertible_to_flux_sed_type:
            raise ValueError(
                "Missing energy range definition, cannot convert to SED type 'eflux'."
            )

        energy_min = self.energy_axis.edges[:-1]
        energy_max = self.energy_axis.edges[1:]
        result = self.reference_spectral_model.energy_flux(energy_min, energy_max)
        return result[self._expand_slice]

    @property
    def dnde(self):
        """Return differential flux (dnde) SED values."""
        return self._use_center_as_labels(self.norm * self.dnde_ref)

    @property
    def dnde_err(self):
        """Return differential flux (dnde) SED errors."""
        return self._use_center_as_labels(self.norm_err * self.dnde_ref)

    @property
    def dnde_errn(self):
        """Return differential flux (dnde) SED negative errors."""
        return self._use_center_as_labels(self.norm_errn * self.dnde_ref)

    @property
    def dnde_errp(self):
        """Return differential flux (dnde) SED positive errors."""
        return self._use_center_as_labels(self.norm_errp * self.dnde_ref)

    @property
    def dnde_ul(self):
        """Return differential flux (dnde) SED upper limit."""
        return self._use_center_as_labels(self.norm_ul * self.dnde_ref)

    @property
    def e2dnde(self):
        """Return differential energy flux (e2dnde) SED values."""
        return self._use_center_as_labels(self.norm * self.e2dnde_ref)

    @property
    def e2dnde_err(self):
        """Return differential energy flux (e2dnde) SED errors."""
        return self._use_center_as_labels(self.norm_err * self.e2dnde_ref)

    @property
    def e2dnde_errn(self):
        """Return differential energy flux (e2dnde) SED negative errors."""
        return self._use_center_as_labels(self.norm_errn * self.e2dnde_ref)

    @property
    def e2dnde_errp(self):
        """Return differential energy flux (e2dnde) SED positive errors."""
        return self._use_center_as_labels(self.norm_errp * self.e2dnde_ref)

    @property
    def e2dnde_ul(self):
        """Return differential energy flux (e2dnde) SED upper limit."""
        return self._use_center_as_labels(self.norm_ul * self.e2dnde_ref)

    @property
    def flux(self):
        """Return integral flux (flux) SED values."""
        return self.norm * self.flux_ref

    @property
    def flux_err(self):
        """Return integral flux (flux) SED values."""
        return self.norm_err * self.flux_ref

    @property
    def flux_errn(self):
        """Return integral flux (flux) SED negative errors."""
        return self.norm_errn * self.flux_ref

    @property
    def flux_errp(self):
        """Return integral flux (flux) SED positive errors."""
        return self.norm_errp * self.flux_ref

    @property
    def flux_ul(self):
        """Return integral flux (flux) SED upper limits."""
        return self.norm_ul * self.flux_ref

    @property
    def flux_sensitivity(self):
        """Sensitivity given as the flux for which the significance is ``self.meta["n_sigma_sensitivity]``."""
        return self.norm_sensitivity * self.flux_ref

    @property
    def eflux(self):
        """Return energy flux (eflux) SED values."""
        return self.norm * self.eflux_ref

    @property
    def eflux_err(self):
        """Return energy flux (eflux) SED errors."""
        return self.norm_err * self.eflux_ref

    @property
    def eflux_errn(self):
        """Return energy flux (eflux) SED negative errors."""
        return self.norm_errn * self.eflux_ref

    @property
    def eflux_errp(self):
        """Return energy flux (eflux) SED positive errors."""
        return self.norm_errp * self.eflux_ref

    @property
    def eflux_ul(self):
        """Return energy flux (eflux) SED upper limits."""
        return self.norm_ul * self.eflux_ref

    def _filter_convergence_failure(self, some_map):
        """Put NaN where pixels did not converge."""
        if not self._filter_success_nan:
            return some_map

        if not self.has_success:
            return some_map

        if self.success.data.shape == some_map.data.shape:
            new_map = some_map.copy()
            new_map.data[~self.success.data] = np.nan
        else:
            mask_nan = np.ones(self.success.data.shape)
            mask_nan[~self.success.data] = np.nan
            new_map = some_map * np.expand_dims(mask_nan, 2)
            new_map.data = new_map.data.astype(some_map.data.dtype)
        return new_map

    def get_flux_points(self, position=None):
        """Extract flux point at a given position.

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            Position where the flux points are extracted.

        Returns
        -------
        flux_points : `~gammapy.estimators.FluxPoints`
            Flux points object.
        """
        from gammapy.estimators import FluxPoints

        if position is None:
            position = self.geom.center_skydir

        data = {}

        for name in self._data:
            m = getattr(self, name)
            if m.data.dtype is np.dtype(bool):
                data[name] = m.to_region_nd_map(
                    region=position, method="nearest", func=np.any
                )
            else:
                data[name] = m.to_region_nd_map(region=position, method="nearest")

        return FluxPoints(
            data,
            reference_model=self.reference_model,
            meta=self.meta.copy(),
            gti=self.gti,
        )

    def to_maps(self, sed_type=None):
        """Return maps in a given SED type.

        Parameters
        ----------
        sed_type : {"likelihood", "dnde", "e2dnde", "flux", "eflux"}, optional
            SED type to convert to. If None, set to `Likelihood`. Default is None.

        Returns
        -------
        maps : `Maps`
            Maps object containing the requested maps.
        """
        maps = Maps()

        if sed_type is None:
            sed_type = self.sed_type_init

        for quantity in self.all_quantities(sed_type=sed_type):
            m = getattr(self, quantity, None)
            if m is not None:
                maps[quantity] = m

        return maps

    @classmethod
    def from_stack(cls, maps, axis, meta=None):
        """Create flux points by stacking list of flux points.

        The first `FluxPoints` object in the list is taken as a reference to infer
        column names and units for the stacked object.

        Parameters
        ----------
        maps : list of `FluxMaps`
            List of maps to stack.
        axis : `MapAxis`
            New axis to create.
        meta : dict, optional
            Metadata of the resulting flux points. Default is None.

        Returns
        -------
        flux_maps : `FluxMaps`
            Stacked flux maps along axis.
        """
        reference = maps[0]
        data = {}

        for quantity in reference.available_quantities:
            data[quantity] = Map.from_stack(
                [_._data[quantity] for _ in maps], axis=axis
            )

        if meta is None:
            meta = reference.meta.copy()

        gtis = [_.gti for _ in maps if _.gti]

        if gtis:
            gti = GTI.from_stack(gtis)
        else:
            gti = None

        return cls(
            data=data, reference_model=reference.reference_model, meta=meta, gti=gti
        )

    def iter_by_axis(self, axis_name):
        """Create a set of FluxMaps by splitting along an axis.

        Parameters
        ----------
        axis_name : str
             Name of the axis to split on.

        Returns
        -------
        flux_maps : `FluxMap`
            FluxMap iteration.

        """
        split_maps = {}
        axis = self.geom.axes[axis_name]
        gti = self.gti

        for amap in self.available_quantities:
            split_maps[amap] = list(getattr(self, amap).iter_by_axis(axis_name))

        for idx in range(axis.nbin):
            maps = {}
            for amap in self.available_quantities:
                maps[amap] = split_maps[amap][idx]
                if isinstance(axis, TimeMapAxis):
                    gti = self.gti.select_time([axis.time_min[idx], axis.time_max[idx]])

            yield self.__class__.from_maps(
                maps=maps,
                sed_type=self.sed_type_init,
                reference_model=self.reference_model,
                gti=gti,
                meta=self.meta,
            )

    @classmethod
    def from_maps(cls, maps, sed_type=None, reference_model=None, gti=None, meta=None):
        """Create FluxMaps from a dictionary of maps.

        Parameters
        ----------
        maps : `Maps`
            Maps object containing the input maps.
        sed_type : str, optional
            SED type of the input maps. If None, set to "likelihood". Default is None.
        reference_model : `~gammapy.modeling.models.SkyModel`, optional
            Reference model to use for conversions.
            If None, a model consisting of a point source with a power
            law spectrum of index 2 is assumed. Default is None.
        gti : `~gammapy.data.GTI`, optional
            Maps GTI information. Default is None.
        meta : `dict`
            Meta dictionary.

        Returns
        -------
        flux_maps : `~gammapy.estimators.FluxMaps`
            Flux maps object.
        """
        if sed_type is None:
            sed_type = cls._guess_sed_type(maps.keys())

        if sed_type is None:
            raise ValueError("Specifying the SED type is required")

        cls._validate_data(data=maps, sed_type=sed_type)

        if sed_type == "likelihood":
            return cls(data=maps, reference_model=reference_model, gti=gti, meta=meta)

        if reference_model is None:
            log.warning(
                "No reference model set for FluxMaps. Assuming point source with E^-2 spectrum."
            )
            reference_model = cls.reference_model_default
        elif isinstance(reference_model, SpectralModel):
            reference_model = SkyModel(reference_model)

        map_ref = maps[sed_type]
        energy_axis = map_ref.geom.axes["energy"]

        with np.errstate(invalid="ignore", divide="ignore"):
            fluxes = reference_model.spectral_model.reference_fluxes(
                energy_axis=energy_axis
            )

        # TODO: handle reshaping in MapAxis
        factor = fluxes[f"ref_{sed_type}"].to(map_ref.unit)[cls._expand_slice]

        data = {}
        data["norm"] = map_ref / factor

        for key in OPTIONAL_QUANTITIES[sed_type]:
            if key in maps:
                norm_type = key.replace(sed_type, "norm")
                data[norm_type] = maps[key] / factor

        # We add the remaining maps
        for key in OPTIONAL_QUANTITIES_COMMON:
            if key in maps:
                data[key] = maps[key]

        return cls(data=data, reference_model=reference_model, gti=gti, meta=meta)

    def to_hdulist(self, sed_type=None, hdu_bands=None):
        """Convert flux map to list of HDUs.

        For now, one cannot export the reference model.

        Parameters
        ----------
        sed_type : str, optional
           SED type to convert to. If None, set to "likelihood". Default is None.
        hdu_bands : str, optional
            Name of the HDU with the BANDS table. Default is 'BANDS'
            If set to None, each map will have its own hdu_band. Default is None.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
            Map dataset list of HDUs.
        """
        if sed_type is None:
            sed_type = self.sed_type_init

        exclude_primary = slice(1, None)

        hdu_primary = fits.PrimaryHDU()
        hdu_primary.header["SED_TYPE"] = sed_type
        hdulist = fits.HDUList([hdu_primary])

        maps = self.to_maps(sed_type=sed_type)
        hdulist.extend(maps.to_hdulist(hdu_bands=hdu_bands)[exclude_primary])

        if self.gti:
            hdu = self.gti.to_table_hdu(format="gadf")
            hdulist.append(hdu)

        return hdulist

    @classmethod
    def from_hdulist(cls, hdulist, hdu_bands=None, sed_type=None, checksum=False):
        """Create flux map dataset from list of HDUs.

        Parameters
        ----------
        hdulist : `~astropy.io.fits.HDUList`
            List of HDUs.
        hdu_bands : str, optional
            Name of the HDU with the BANDS table. Default is 'BANDS'
            If set to None, each map should have its own hdu_band. Default is None.
        sed_type : {"dnde", "flux", "e2dnde", "eflux", "likelihood"}, optional
            Sed type. Default is None.

        Returns
        -------
        flux_maps : `~gammapy.estimators.FluxMaps`
            Flux maps object.
        """
        maps = Maps.from_hdulist(hdulist=hdulist, hdu_bands=hdu_bands)

        if sed_type is None:
            sed_type = hdulist[0].header.get("SED_TYPE", None)

        filename = hdulist[0].header.get("MODEL", None)

        if filename:
            reference_model = Models.read(filename, checksum=checksum)[0]
        else:
            reference_model = None

        if "GTI" in hdulist:
            gti = GTI.from_table_hdu(hdulist["GTI"])
        else:
            gti = None

        return cls.from_maps(
            maps=maps, sed_type=sed_type, reference_model=reference_model, gti=gti
        )

    def write(
        self,
        filename,
        filename_model=None,
        overwrite=False,
        sed_type=None,
        checksum=False,
    ):
        """Write flux map to file.

        Parameters
        ----------
        filename : str
            Filename to write to.
        filename_model : str
            Filename of the model (yaml format).
            If None, keep string before '.' and add '_model.yaml' suffix.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        sed_type : str, optional
            Sed type to convert to. If None, set to "likelihood". Default is None.
        checksum : bool, optional
            When True adds both DATASUM and CHECKSUM cards to the headers written to the file.
            Default is False.
        """
        if sed_type is None:
            sed_type = self.sed_type_init

        filename = make_path(filename)

        if filename_model is None:
            name_string = filename.as_posix()
            for suffix in filename.suffixes:
                name_string.replace(suffix, "")
            filename_model = name_string + "_model.yaml"

        filename_model = make_path(filename_model)

        hdulist = self.to_hdulist(sed_type)

        models = Models(self.reference_model)
        models.write(filename_model, overwrite=overwrite, write_covariance=False)
        hdulist[0].header["MODEL"] = filename_model.as_posix()

        hdulist.writeto(filename, overwrite=overwrite)

    @classmethod
    def read(cls, filename, checksum=False):
        """Read map dataset from file.

        Parameters
        ----------
        filename : str
            Filename to read from.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        flux_maps : `~gammapy.estimators.FluxMaps`
            Flux maps object.
        """
        with fits.open(
            str(make_path(filename)), memmap=False, checksum=checksum
        ) as hdulist:
            return cls.from_hdulist(hdulist, checksum=checksum)

    def copy(self, reference_model=None):
        """Deep copy.

        Parameters
        ----------
        reference_model : `~gammapy.modeling.models.SkyModel`, optional
            The reference model to use for conversions. If None, the original model is copied.
            Flux maps have been obtained for a specific reference model.
            Changing it will change the fluxes. Handle with care.

        Returns
        -------
        flux_maps : `~gammapy.estimators.FluxMaps`
            Copied flux maps object.

        """
        new = deepcopy(self)
        if reference_model is not None:
            new._reference_model = reference_model.copy()
            log.warning(
                "Changing the reference model will change the fluxes. Handle with care."
            )
        return new

    def slice_by_idx(self, slices):
        """Slice flux maps by index.

        Parameters
        ----------
        slices : dict
            Dictionary of axes names and integers or `slice` object pairs. Contains one
            element for each non-spatial dimension. For integer indexing the
            corresponding axes is dropped from the map. Axes not specified in the
            dict are kept unchanged.

        Returns
        -------
        flux_maps : `FluxMaps`
            Sliced flux maps object.

        Examples
        --------
        >>> from gammapy.estimators import FluxPoints
        >>> import astropy.units as u
        >>> fp = FluxPoints.read("$GAMMAPY_DATA/estimators/crab_hess_fp/crab_hess_fp.fits")
        >>> slices = {"energy": slice(0, 2)}
        >>> sliced = fp.slice_by_idx(slices)
        """
        data = {}

        for key, item in self._data.items():
            data[key] = item.slice_by_idx(slices)

        return self.__class__(
            data=data,
            reference_model=self.reference_model,
            meta=self.meta.copy(),
            gti=self.gti,
        )

    def slice_by_coord(self, slices):
        """Slice flux maps by coordinate values.

        Parameters
        ----------
        slices : dict
            Dictionary of axes names and `astropy.Quantity` or `astropy.Time` or `slice` object pairs.
            Contains one element for each non-spatial dimension. For integer indexing the
            corresponding axes is dropped from the map. Axes not specified in the
            dict are kept unchanged.


        Returns
        -------
        flux_maps : `FluxMaps`
            Sliced flux maps object.

        Examples
        --------
        >>> from gammapy.estimators import FluxPoints
        >>> import astropy.units as u
        >>> lc_1d = FluxPoints.read("$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits")
        >>> slices = {"time": slice(2035.93*u.day, 2036.05*u.day)}
        >>> sliced = lc_1d.slice_by_coord(slices)
        """
        idx_intervals = []

        for key, interval in zip(slices.keys(), slices.values()):
            axis = self.geom.axes[key]

            group = axis.group_table([interval.start, interval.stop])

            is_normal = group["bin_type"] == "normal   "
            group = group[is_normal]

            idx_intervals.append(
                slice(int(group["idx_min"][0]), int(group["idx_max"][0] + 1))
            )

        return self.slice_by_idx(dict(zip(slices.keys(), idx_intervals)))

    def slice_by_time(self, time_min, time_max):
        """Slice flux maps by coordinate values along the time axis.

        Parameters
        ----------
        time_min, time_max : `~astropy.time.Time`
            Time bounds used to slice the flux map.

        Returns
        -------
        flux_maps : `FluxMaps`
            Sliced flux maps object.

        Examples
        --------
        >>> from gammapy.estimators import FluxPoints
        >>> import astropy.units as u
        >>> lc_1d = FluxPoints.read("$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits")
        >>> sliced = lc_1d.slice_by_time(time_min=2035.93*u.day, time_max=2036.05*u.day)
        """
        time_slice = slice(time_min, time_max)

        return self.slice_by_coord({"time": time_slice})

    def slice_by_energy(self, energy_min, energy_max):
        """Slice flux maps by coordinate values along the energy axis.

        Parameters
        ----------
        energy_min, energy_max : `~astropy.units.Quantity`
            Energy bounds used to slice the flux map.

        Returns
        -------
        flux_maps : `FluxMaps`
            Sliced flux maps object.

        Examples
        --------
        >>> from gammapy.estimators import FluxPoints
        >>> import astropy.units as u
        >>> fp = FluxPoints.read("$GAMMAPY_DATA/estimators/crab_hess_fp/crab_hess_fp.fits")
        >>> sliced = fp.slice_by_energy(energy_min=2*u.TeV, energy_max=10*u.TeV)
        """
        energy_slice = slice(energy_min, energy_max)

        return self.slice_by_coord({"energy": energy_slice})

    # TODO: should we allow this?
    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        str_ = f"{self.__class__.__name__}\n"
        str_ += "-" * len(self.__class__.__name__)
        str_ += "\n\n"
        str_ += "\t" + f"geom                   : {self.geom.__class__.__name__}\n"
        str_ += "\t" + f"axes                   : {self.geom.axes_names}\n"
        str_ += "\t" + f"shape                  : {self.geom.data_shape[::-1]}\n"
        str_ += "\t" + f"quantities             : {list(self.available_quantities)}\n"
        str_ += (
            "\t" + f"ref. model             : {self.reference_spectral_model.tag[-1]}\n"
        )
        str_ += "\t" + f"n_sigma                : {self.n_sigma}\n"
        str_ += "\t" + f"n_sigma_ul             : {self.n_sigma_ul}\n"
        str_ += "\t" + f"sqrt_ts_threshold_ul   : {self.sqrt_ts_threshold_ul}\n"
        str_ += "\t" + f"sed type init          : {self.sed_type_init}\n"
        return str_.expandtabs(tabsize=2)



r"""
.. _powerlaw-temporal-model:

PowerLaw temporal model
=======================

This model parametrises a power-law time model.

.. math:: F(t) = \left( \frac{t - t_{\rm{ref}}}{t_0} \right)^\alpha

"""

# %%
# Example plot
# ------------
# Here is an example plot of the model:

from astropy import units as u
from astropy.time import Time
import matplotlib.pyplot as plt
from gammapy.modeling.models import (
    Models,
    PowerLawSpectralModel,
    PowerLawTemporalModel,
    SkyModel,
)

time_range = [Time.now(), Time.now() + 2 * u.d]
pl_model = PowerLawTemporalModel(alpha=-2.0, t_ref=(time_range[0].mjd - 0.1) * u.d)
pl_model.plot(time_range)
plt.grid(which="both")
plt.yscale("log")

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(
    spectral_model=PowerLawSpectralModel(),
    temporal_model=pl_model,
    name="powerlaw-model",
)
models = Models([model])

print(models.to_yaml())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from gammapy.modeling import Fit, Parameter, Covariance, FitResult, OptimizeResult
from gammapy.stats.utils import sigma_to_ts

__all__ = ["select_nested_models", "NestedModelSelection"]


class NestedModelSelection:
    """Compute the test statistic (TS) between two nested hypothesis.

    The null hypothesis is the minimal one, for which a set of parameters
    are frozen to given values. The model is updated to the alternative hypothesis
    if there is a significant improvement (larger than the given threshold).

    Parameters
    ----------
    parameters : `~gammapy.modeling.Parameters` or list of `~gammapy.modeling.Parameter`
        List of parameters frozen for the null hypothesis but free for the test hypothesis.
    null_values : list of float or `~gammapy.modeling.Parameters`
        Values of the parameters frozen for the null hypothesis.
        If a `Parameters` object or a list of `Parameters` is given
        the null hypothesis follows the values of these parameters,
        so this tests linked parameters versus unliked.
    n_sigma : float
        Threshold in number of sigma to switch from the null hypothesis
        to the alternative one. Default is 2.
        The TS is converted to sigma assuming that the Wilks' theorem is verified.
    n_free_parameters : int
        Number of free parameters to consider between the two hypothesis
        in order to estimate the `ts_threshold` from the `n_sigma` threshold.
        Default is len(parameters).
    fit : `Fit`
        Fit instance specifying the backend and fit options.
    """

    def __init__(
        self, parameters, null_values, n_sigma=2, n_free_parameters=None, fit=None
    ):
        self.parameters = parameters
        self.null_values = null_values
        self.n_sigma = n_sigma

        if n_free_parameters is None:
            n_free_parameters = len(parameters)
        self.n_free_parameters = n_free_parameters

        if fit is None:
            fit = Fit()
            minuit_opts = {"tol": 0.1, "strategy": 1}
            fit.backend = "minuit"
            fit.optimize_opts = minuit_opts
        self.fit = fit

    @property
    def ts_threshold(self):
        """Threshold value in TS corresponding to `n_sigma`.

        This assumes that the TS follows a chi squared distribution
        with a number of degree of freedom equal to `n_free_parameters`.
        """
        return np.sign(self.n_sigma) * sigma_to_ts(self.n_sigma, self.n_free_parameters)

    def ts_known_bkg(self, datasets):
        """Perform the alternative hypothesis testing assuming known background (all parameters frozen).

        This implicitly assumes that the alternative model is a good representation of the true model.
        If the assumption is true, the ``ts_known_bkg`` should tend to the ``ts_asimov`` (deviation would
        indicate a bad fit of the data).
        Deviations between ``ts`` and ``frozen_ts`` can be used to identify potential sources of confusion
        depending on which parameters are let free for the ``ts`` computation
         (for example considering diffuse background or nearby source).
        """
        stat = datasets.stat_sum()
        cache = self._apply_null_hypothesis(datasets)
        stat_null = datasets.stat_sum()
        self._restore_status(datasets, cache)
        return stat_null - stat

    def ts_asimov(self, datasets):
        """Perform the alternative hypothesis testing in the Asimov dataset.

        The Asimov dataset is defined by counts=npred such as the non-null model is the true model.
        """
        asimov_datasets = datasets._to_asimov_datasets()
        return self.ts_known_bkg(asimov_datasets)

    def ts(self, datasets):
        """Perform the alternative hypothesis testing."""
        return self.run(datasets, apply_selection=False)["ts"]

    def run(self, datasets, apply_selection=True):
        """Perform the alternative hypothesis testing and apply model selection.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.
        apply_selection : bool, optional
            Apply or not the model selection. Default is True.

        Returns
        -------
        result : dict
            Dictionary with the TS of the best fit value compared to the null hypothesis
            and fit results for the two hypotheses. Entries are:

                * "ts" : fit statistic difference with null hypothesis
                * "fit_results" : results for the best fit
                * "fit_results_null" : fit results for the null hypothesis
        """
        for p in self.parameters:
            p.frozen = False
        fit_results = self.fit.run(datasets)
        stat = datasets.stat_sum()

        cache = self._apply_null_hypothesis(datasets)

        if len(datasets.models.parameters.free_parameters) > 0:
            fit_results_null = self.fit.run(datasets)
        else:
            fit_results_null = FitResult(
                OptimizeResult(
                    models=datasets.models.copy(),
                    nfev=0,
                    total_stat=datasets.stat_sum(),
                    trace=None,
                    backend=None,
                    method=None,
                    success=None,
                    message=None,
                )
            )
        stat_null = datasets.stat_sum()

        ts = stat_null - stat
        if not apply_selection or ts > self.ts_threshold:
            # restore default model if preferred against null hypothesis or if selection is ignored
            self._restore_status(datasets, cache)
        return dict(
            ts=ts,
            fit_results=fit_results,
            fit_results_null=fit_results_null,
        )

    def _apply_null_hypothesis(self, datasets):
        cache = dict()
        cache["object"] = [p.__dict__ for p in datasets.models.parameters]
        cache["values"] = [p.value for p in datasets.models.parameters]
        cache["error"] = [p.error for p in datasets.models.parameters]
        for p, val in zip(self.parameters, self.null_values):
            if isinstance(val, Parameter):
                p.__dict__ = val.__dict__
            else:
                p.value = val
                p.frozen = True
        cache["covar"] = Covariance(
            datasets.models.parameters, datasets.models.covariance.data
        )
        return cache

    def _restore_status(self, datasets, cache):
        """Restore parameters to given cached objects and values."""
        for p in self.parameters:
            p.frozen = False
        for kp, p in enumerate(datasets.models.parameters):
            p.__dict__ = cache["object"][kp]
            p.value = cache["values"][kp]
            p.error = cache["error"][kp]
        datasets._covariance = cache["covar"]


def select_nested_models(
    datasets, parameters, null_values, n_sigma=2, n_free_parameters=None, fit=None
):
    """Compute the test statistic (TS) between two nested hypotheses.

    This function evaluates whether adding one or more parameters (alternative hypothesis)
    provides a statistically significant improvement (larger than the given threshold)
    over a simpler model (null hypothesis), where those parameters are fixed to given values.

    The model is updated to the alternative hypothesis if there is a significant
    improvement (larger than the given threshold, ``n_sigma``).

    Parameters
    ----------
    datasets : `~gammapy.datasets.Datasets`
        Datasets.
    parameters : `~gammapy.modeling.Parameters` or list of `~gammapy.modeling.Parameter`
        List of parameters frozen for the null hypothesis but free for the test hypothesis.
    null_values : list of float or `~gammapy.modeling.Parameters`
        Values of the parameters frozen under the null hypothesis.
        If a `Parameters` object or a list of `Parameters` is provided,
        the null hypothesis assumes these values.
        This allows testing of linked versus unlinked parameters.
    n_sigma : float, optional
        Threshold in number of sigma for rejecting the null hypothesis in favour of
        the alternative. Default is 2.
        The TS is converted to sigma assuming that the Wilks' theorem is verified.
    n_free_parameters : int, optional
        Number of free parameters to consider between the two hypotheses
        when estimating the `ts_threshold` from the ``n_sigma`` threshold.
        Default is ``len(parameters)``.
    fit : `Fit`, optional
        Fit instance specifying the backend and fit options. Default is None, which utilises
        the "minuit" backend with tol=0.1 and strategy=1.

    Returns
    -------
    result : dict
        Dictionary with the TS of the best fit value compared to the null hypothesis
        and fit results for the two hypotheses. Entries are:

            * "ts" : fit statistic difference with null hypothesis
            * "fit_results" : results for the best fit
            * "fit_results_null" : fit results for the null hypothesis

    Notes
    -----
    This is useful for testing the significance of adding model components, such as
    a spectral cutoff or a source detection. It assumes the two models are *nested* —
    the null hypothesis is a special case of the alternative with fixed parameters.

    See below for two simple examples and :ref:`compute_upper_lims` for a more in-depth example.

    Examples
    --------
    .. testcode::

        from gammapy.modeling.selection import select_nested_models
        from gammapy.datasets import Datasets, SpectrumDatasetOnOff
        from gammapy.modeling.models import SkyModel

        # Test if cutoff is significant

        dataset = SpectrumDatasetOnOff.read("$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits")
        datasets = Datasets(dataset)
        model = SkyModel.create(spectral_model="ecpl", spatial_model="point", name='hess')
        datasets.models = model
        # If there is no cutoff (null hypothesis) then the spectral cutoff (`lambda`) is zero
        result = select_nested_models(datasets,
                                      parameters=[model.spectral_model.lambda_],
                                      null_values=[0],
                                      )

        # Test if source is significant

        filename = "$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_datasets.yaml"
        filename_models = "$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_models.yaml"
        fermi_datasets = Datasets.read(filename=filename, filename_models=filename_models)
        model = fermi_datasets.models["Crab Nebula"]
        # Number of parameters previously fit for the source of interest
        n_free_parameters = len(model.parameters.free_parameters)
        # Freeze spatial parameters to ensure another weaker source does not move from its position
        # to replace the source of interest during the null hypothesis test.
        # (with all parameters free you test N vs. N+1 models and not the detection of a specific source.)
        fermi_datasets.models.freeze(model_type='spatial')
        # Null hypothesis value for the amplitude should be 0, i.e. the flux of the source is 0
        results = select_nested_models(fermi_datasets,
                                      parameters=[model.spectral_model.amplitude],
                                      null_values=[0],
                                      n_free_parameters=n_free_parameters,
                                      n_sigma=4,
                                      )
    """
    test = NestedModelSelection(
        parameters, null_values, n_sigma, n_free_parameters, fit
    )
    return test.run(datasets)



"""
Estimators
==========

This tutorial provides an overview of the ``Estimator`` API. All estimators live in the
`gammapy.estimators` sub-module, offering a range of algorithms and classes for high-level flux and
significance estimation. This is accomplished through a common functionality allowing the estimation of
flux points, light curves, flux maps and profiles via a common API.



Key Features
------------

-  **Hypothesis Testing**: Estimations are based on testing a reference model
   against a null hypothesis, deriving flux and significance values.

-  **Estimation via Two Methods**:

   -   **Model Fitting (Forward Folding)**: Refit the flux of a model component
       within specified energy, time, or spatial regions.
   -   **Excess Calculation (Backward Folding)**: Use the analytical solution by Li and Ma
       for significance based on excess counts, currently available in `~gammapy.estimators.ExcessMapEstimator`.

For further information on these details please refer to :doc:`/user-guide/estimators`.

The setup
---------

"""

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u
from IPython.display import display
from gammapy.datasets import SpectrumDatasetOnOff, Datasets, MapDataset
from gammapy.estimators import (
    FluxPointsEstimator,
    ExcessMapEstimator,
    FluxPoints,
)
from gammapy.modeling import Fit, Parameter
from gammapy.modeling.models import SkyModel, PowerLawSpectralModel
from gammapy.utils.scripts import make_path


######################################################################
# Flux Points Estimation
# ----------------------
#
# We start with a simple example for flux points estimation taking multiple datasets into account.
# In this section we show the steps to estimate the flux points.
# First we read the pre-computed datasets from `$GAMMAPY_DATA`.
#

datasets = Datasets()
path = make_path("$GAMMAPY_DATA/joint-crab/spectra/hess/")

for filename in path.glob("pha_obs*.fits"):
    dataset = SpectrumDatasetOnOff.read(filename)
    datasets.append(dataset)

######################################################################
# Next we define a spectral model and set it on the datasets:
#

pwl = PowerLawSpectralModel(index=2.7, amplitude="5e-11  cm-2 s-1 TeV-1")
datasets.models = SkyModel(spectral_model=pwl, name="crab")

######################################################################
# Before using the estimators, it is necessary to first ensure that the model is properly
# fitted. This applies to all scenarios, including light curve estimation. To optimize the
# model parameters to best fit the data we utilise the following:
#

fit = Fit()
fit_result = fit.optimize(datasets=datasets)
print(fit_result)

######################################################################
# A fully configured Flux Points Estimation
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The `~gammapy.estimators.FluxPointsEstimator` estimates flux points for a given list of datasets,
# energies and spectral model. The most simple way to call the estimator is by defining both
# the name of the ``source`` and its ``energy_edges``.
# Here we prepare a full configuration of the flux point estimation.
# Firstly we define the ``backend`` for the fit:
#

fit = Fit(
    optimize_opts={"backend": "minuit"},
    confidence_opts={"backend": "scipy"},
)

######################################################################
# Define the fully configured flux points estimator:
#

energy_edges = np.geomspace(0.7, 100, 9) * u.TeV
norm = Parameter(name="norm", value=1.0)

fp_estimator = FluxPointsEstimator(
    source="crab",
    energy_edges=energy_edges,
    n_sigma=1,
    n_sigma_ul=2,
    selection_optional="all",
    fit=fit,
    norm=norm,
)

######################################################################
# The ``norm`` parameter can be adjusted in a few different ways. For example, we can change its
# minimum and maximum values that it scans over, as follows.
#

fp_estimator.norm.scan_min = 0.1
fp_estimator.norm.scan_max = 10

######################################################################
# Note: The default scan range of the norm parameter is between 0.2 to 5. In case the upper
# limit values lie outside this range, nan values will be returned. It may thus be useful to
# increase this range, specially for the computation of upper limits from weak sources.
#
# The various quantities utilised in this tutorial are described here:
#
# -  ``source``: which source from the model to compute the flux points for
# -  ``energy_edges``: edges of the flux points energy bins
# -  ``n_sigma``: number of sigma for the flux error
# -  ``n_sigma_ul``: the number of sigma for the flux upper limits
# -  ``selection_optional``: what additional maps to compute
# -  ``fit``: the fit instance (as defined above)
# -  ``reoptimize``: whether to reoptimize the flux points with other model parameters, aside from the ``norm``
# -  ``norm``: normalisation parameter for the fit
#
# **Important note**: the output ``energy_edges`` are taken from the parent dataset energy bins,
# selecting the bins closest to the requested ``energy_edges``. To match the input bins directly,
# specific binning must be defined based on the parent dataset geometry. This could be done in the following way:
# ``energy_edges = datasets[0].geoms["geom"].axes["energy"].downsample(factor=5).edges``
#


fp_result = fp_estimator.run(datasets=datasets)

######################################################################
# Accessing and visualising the results
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

print(fp_result)

######################################################################
# We can specify the SED type to plot:
#
fp_result.plot(sed_type="dnde")
plt.show()

######################################################################
# We can also access
# the quantities names through ``fp_result.available_quantities``.
# Here we show how you can plot a different plot type and define the axes units,
# we also overlay the TS profile.

ax = plt.subplot()
ax.xaxis.set_units(u.eV)
ax.yaxis.set_units(u.Unit("TeV cm-2 s-1"))
fp_result.plot(ax=ax, sed_type="e2dnde", color="tab:orange")
fp_result.plot_ts_profiles(sed_type="e2dnde")
plt.show()

######################################################################
# The actual data members are N-dimensional `~gammapy.maps.region.ndmap.RegionNDMap` objects. So you can
# also plot them:

print(type(fp_result.dnde))

######################################################################
#
fp_result.dnde.plot()
plt.show()

######################################################################
# From the above, we can see that we access to many quantities.


######################################################################
# Access the data:

print(fp_result.e2dnde.quantity.to("TeV cm-2 s-1"))

######################################################################
#
print(fp_result.dnde.quantity.shape)

######################################################################
#
print(fp_result.dnde.quantity[:, 0, 0])

######################################################################
# Or even extract an energy range:

fp_result.dnde.slice_by_idx({"energy": slice(3, 10)})


######################################################################
# A note on the internal representation
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The result contains a reference spectral model, which defines the spectral shape.
# Typically, it is the best fit model:

print(fp_result.reference_model)

######################################################################
# `~gammapy.estimators.FluxPoints` are the represented by the "norm" scaling factor with
# respect to the reference model:

fp_result.norm.plot()
plt.show()

######################################################################
# Dataset specific quantities ("counts like")
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# While the flux estimate and associated errors are common to all datasets,
# the result also stores some dataset specific quantities, which can be useful
# for debugging.
# Here we remind the user of the meaning of the forthcoming quantities:
#
# -  ``counts``: predicted counts from the null hypothesis,
# -  ``npred``: predicted number of counts from best fit hypothesis,
# -  ``npred_excess``: predicted number of excess counts from best fit hypothesis.
#
# The `~gammapy.maps.region.ndmap.RegionNDMap` allows for plotting of multidimensional data
# as well, by specifying the primary ``axis_name``:


fp_result.counts.plot(axis_name="energy")
plt.show()

######################################################################
#
fp_result.npred.plot(axis_name="energy")
plt.show()

######################################################################
#
fp_result.npred_excess.plot(axis_name="energy")
plt.show()

######################################################################
# Table conversion
# ~~~~~~~~~~~~~~~~
#
# Flux points can be converted to tables:
#

table = fp_result.to_table(sed_type="flux", format="gadf-sed")
display(table)

######################################################################
#
table = fp_result.to_table(sed_type="likelihood", format="gadf-sed", formatted=True)
display(table)


######################################################################
# Common API
# ----------
# In `GAMMAPY_DATA` we have access to other `~gammapy.estimators.FluxPoints` objects
# which have been created utilising the above method. Here we read the PKS 2155-304 light curve
# and create a `~gammapy.estimators.FluxMaps` object and show the data structure of such objects.
# We emphasize that these follow a very similar structure.
#

######################################################################
# Load the light curve for the PKS 2155-304 as a `~gammapy.estimators.FluxPoints` object.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#

lightcurve = FluxPoints.read(
    "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits", format="lightcurve"
)

display(lightcurve.available_quantities)


######################################################################
# Create a `~gammapy.estimators.FluxMaps` object through one of the estimators.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#

dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
estimator = ExcessMapEstimator(correlation_radius="0.1 deg")
result = estimator.run(dataset)
display(result)

######################################################################
#
display(result.available_quantities)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Instrument response functions (IRFs)."""
from gammapy.utils.registry import Registry
from .background import Background2D, Background3D, BackgroundIRF
from .core import IRF, FoVAlignment, IRFMap
from .edisp import EDispKernel, EDispKernelMap, EDispMap, EnergyDispersion2D
from .effective_area import EffectiveAreaTable2D
from .io import load_irf_dict_from_file
from .psf import (
    PSF3D,
    EnergyDependentMultiGaussPSF,
    ParametricPSF,
    PSFKernel,
    PSFKing,
    PSFMap,
    RecoPSFMap,
)
from .rad_max import RadMax2D

__all__ = [
    "Background2D",
    "Background3D",
    "BackgroundIRF",
    "EDispKernel",
    "EDispKernelMap",
    "EDispMap",
    "EffectiveAreaTable2D",
    "EnergyDependentMultiGaussPSF",
    "EnergyDispersion2D",
    "FoVAlignment",
    "IRF_REGISTRY",
    "IRFMap",
    "IRF",
    "load_irf_dict_from_file",
    "ParametricPSF",
    "PSF3D",
    "PSFKernel",
    "PSFKing",
    "PSFMap",
    "RecoPSFMap",
    "RadMax2D",
]


IRF_REGISTRY = Registry(
    [
        EffectiveAreaTable2D,
        EnergyDispersion2D,
        PSF3D,
        EnergyDependentMultiGaussPSF,
        PSFKing,
        Background3D,
        Background2D,
        PSFMap,
        RecoPSFMap,
        EDispKernelMap,
        RadMax2D,
        EDispMap,
    ]
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
from collections.abc import MutableMapping
from astropy.io import fits
from gammapy.maps import Map
from gammapy.utils.scripts import make_path

__all__ = ["Maps"]


class Maps(MutableMapping):
    """A Dictionary containing Map objects sharing the same geometry.

    This class simplifies handling and I/O of maps collections.

    For maps with different geometries, use a regular dict.
    """

    def __init__(self, **kwargs):
        self._geom = None
        self._data = {}
        for key, value in kwargs.items():
            self.__setitem__(key, value)

    @property
    def geom(self):
        """Map geometry as a `~gammapy.maps.Geom` object."""
        return self._geom

    def __setitem__(self, key, value):
        if value is not None and not isinstance(value, Map):
            raise ValueError(
                f"MapDict can only contain Map objects, got {type(value)} instead."
            )
        # TODO: which loosers criterion to apply? broadcastability?
        else:
            self._geom = value.geom

        self._data[key] = value

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        """Returns the length of MapDict."""
        return len(self._data)

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __repr__(self):
        return f"{type(self).__name__}({self._data})"

    def __str__(self):
        str_ = f"{self.__class__.__name__}\n"
        str_ += "-" * len(self.__class__.__name__) + "\n"
        str_ += "\n"
        str_ += self._geom.__repr__()
        str_ += "\n"
        for name, value in self.items():
            str_ += f"{name} \n"
            str_ += f"\t unit\t : {value.unit} \n"
            str_ += f"\t dtype\t : {value.data.dtype}\n"
            str_ += "\n"
        return str_

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def to_hdulist(self, hdu_bands="BANDS"):
        """Convert map dictionary to list of HDUs.

        Parameters
        ----------
        hdu_bands : str, optional
            Name of the HDU with the BANDS table. If set to None, each map will have its own hdu_band.
            Default is 'BANDS'.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
            Map dataset list of HDUs.
        """
        exclude_primary = slice(1, None)

        hdu_primary = fits.PrimaryHDU()
        hdulist = fits.HDUList([hdu_primary])

        for key, m in self.items():
            hdulist += m.to_hdulist(hdu=key, hdu_bands=hdu_bands)[exclude_primary]

        return hdulist

    @classmethod
    def from_hdulist(cls, hdulist, hdu_bands="BANDS"):
        """Create map dictionary from a HDU list.

        Because FITS keywords are case-insensitive, all key names will return as lower-case.

        Parameters
        ----------
        hdulist : `~astropy.io.fits.HDUList`
            List of HDUs.
        hdu_bands : str, optional
            Name of the HDU with the BANDS table. If set to None, each map should have its own hdu_band.
            Default is 'BANDS'.

        Returns
        -------
        maps : `~gammapy.maps.Maps`
            Maps object.
        """
        maps = cls()

        for hdu in hdulist:
            if hdu.is_image and hdu.data is not None:
                map_name = hdu.name.lower()
                maps[map_name] = Map.from_hdulist(
                    hdulist, hdu=map_name, hdu_bands=hdu_bands
                )
        return maps

    @classmethod
    def read(cls, filename, checksum=False):
        """Read map dictionary from file.

        Because FITS keywords are case-insensitive, all key names will return as lower-case.

        Parameters
        ----------
        filename : str
            Filename to read from.

        Returns
        -------
        maps : `~gammapy.maps.Maps`
            Maps object.
        """
        with fits.open(
            str(make_path(filename)), memmap=False, checksum=checksum
        ) as hdulist:
            return cls.from_hdulist(hdulist)

    def write(self, filename, overwrite=False, checksum=False):
        """Write map dictionary to file.

        Parameters
        ----------
        filename : str
            Filename to write to.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        checksum : bool, optional
            When True adds both DATASUM and CHECKSUM cards to the headers written to the file.
            Default is False.
        """
        filename = make_path(filename)

        hdulist = self.to_hdulist()
        hdulist.writeto(filename, overwrite=overwrite, checksum=checksum)

    @classmethod
    def from_geom(cls, geom, names, kwargs_list=None):
        """Create map dictionary from a geometry.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            The input geometry that will be used by all maps.
        names : list of str
            The list of all map names.
        kwargs_list : list of dict
            the list of arguments to be passed to `~gammapy.maps.Map.from_geom()`.

        Returns
        -------
        maps : `~gammapy.maps.Maps`
            Maps object.
        """
        mapdict = {}

        if kwargs_list is None:
            kwargs_list = [{}] * len(names)

        for name, kwargs in zip(names, kwargs_list):
            mapdict[name] = Map.from_geom(geom, **kwargs)

        return cls(**mapdict)



"""
Computing flux upper limits
===========================

Explore how to compute flux upper limits for a non-detected source.

Prerequisites
-------------

It is advisable to understand the general Gammapy modeling and fitting
framework before proceeding with this notebook, e.g. :doc:`/user-guide/modeling`.

Context
-------

In the study of VHE sources, one often encounters no significant
detection even after long exposures. In that case, it may be useful to
compute flux upper limits (UL) for the said target consistent with the observation.

Proposed approach
-----------------

In this section, we will use an empty observation from the H.E.S.S. DL3 DR1 to understand how
to quantify non-detections. There are two distinct approaches to consider:

- Test for the presence of emission anywhere in a map and compute an integral flux upper limit at
  any position (i.e. UL map).
- Test the presence of emission from a potential source with given position and morphology and compute
  integral and differential UL

"""

######################################################################
# Setup
# -----
#
# As usual, let’s start with some general imports…
#

# %matplotlib inline
import matplotlib.pyplot as plt

import astropy.units as u

from gammapy.datasets import MapDataset, Datasets
from gammapy.estimators import FluxPointsEstimator, ExcessMapEstimator
from gammapy.modeling import select_nested_models
from gammapy.modeling.models import (
    PointSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
    create_crab_spectral_model,
)
from gammapy.visualization import plot_distribution


######################################################################
# Load observation
# ----------------
#
# For computational purposes, we have
# already created a `~gammapy.datasets.MapDataset` from observation id ``20275`` from the
# public H.E.S.S. data release and stored it in ``$GAMMAPY_DATA``
#

dataset = MapDataset.read("$GAMMAPY_DATA/datasets/empty-dl4/empty-dl4.fits.gz")
dataset.peek()
plt.show()

######################################################################
# Create Upper Limit maps
# -----------------------
#
# We will first use the `~gammapy.estimators.ExcessMapEstimator` for a quick check to see if
# there are any potential sources in the field. The ``correlation_radius`` should be around the
# size of the source you are searching for. You may also use the
# `~gammapy.estimators.TSMapEstimator`.

estimator = ExcessMapEstimator(
    sum_over_energy_groups=True,
    selection_optional="all",
    correlate_off=True,
    correlation_radius=0.1 * u.deg,
)

lima_maps = estimator.run(dataset)

significance_map = lima_maps["sqrt_ts"]
excess_map = lima_maps["npred_excess"]

fig, (ax1, ax2) = plt.subplots(
    figsize=(11, 4), subplot_kw={"projection": lima_maps.geom.wcs}, ncols=2
)
ax1.set_title("Significance map")
significance_map.plot(ax=ax1, add_cbar=True)
ax2.set_title("Excess map")
excess_map.plot(ax=ax2, add_cbar=True)
plt.show()

######################################################################
# The significance map looks featureless. We will plot a histogram of the
# significance distribution and fit it with a standard normal.
# Deviations from a standard normal can suggest the presence of gamma-ray sources,
# or can also originate from incorrect modeling of the residual hadronic background.

ax = plt.subplot()
plot_distribution(
    significance_map,
    func="norm",
    ax=ax,
    kwargs_hist={"bins": 50, "range": (-4, 4), "density": True},
)
plt.show()


######################################################################
# We can also see the correlated upper limits at any position in the map. However, it is important to note
# that this is **not** a source UL, as the containment correction is not applied here. Instead, it gives the
# flux upper limits contained within the ``correlation_radius`` at each pixel. This can be useful
# when making quick look plots to search for the presence of new sources with a field - for example
# in the case of alerts from Gravitational Wave detectors.
#

lima_maps.flux_ul.plot(add_cbar=True, cmap="viridis")
plt.show()


###############################################################################
# .. _compute_upper_lims:
#
# Compute upper limits for a source
# ---------------------------------
#
# Now, we address a more specific question.
# Suppose we were expecting a source at a specific position, say the center of our map.
# We’ll attempt to fit a point source at that location to determine whether the signal
# is significant above the background.
#
# For this, we compare two hypotheses:
#
# - Null hypothesis, :math:`H_0` - only the background is present (i.e., no source)
# - Alternative hypothesis, :math:`H_1` - the background plus a point source
#
# The difference in test statistic (TS) between the two cases indicates the
# significance of the alternative hypothesis.
#
# For this purpose, we can utilise the function `~gammapy.modeling.select_nested_models` which
# performs these comparisons internally.
# As the null hypothesis, :math:`H_0`, corresponds to the case of no source, we set the amplitude to 0, in other words
# the source has no flux.
#
# To prevent the fit from converging to unrelated positions, we freeze the spatial parameters.
# Alternatively, you can constrain the parameter ranges to stay within your expected region.


spectral_model = PowerLawSpectralModel()
spatial_model = PointSpatialModel(lon_0=187 * u.deg, lat_0=2.6 * u.deg, frame="icrs")
spatial_model.lat_0.frozen = True
spatial_model.lon_0.frozen = True

sky_model = SkyModel(
    spatial_model=spatial_model, spectral_model=spectral_model, name="test"
)
dataset.models = sky_model
LLR = select_nested_models(
    datasets=Datasets(dataset),
    parameters=[sky_model.parameters["amplitude"]],
    null_values=[0],
)
print(LLR)

######################################################################
# The fitted parameters under the alternative hypothesis:

print(LLR["fit_results"].parameters.to_table())

######################################################################
# The test statistic is ``ts ~ 4.7``. Note that here we have
# only 1 free parameter, the amplitude, and thus, we can assume the simple conversion
# significance = :math:`\sqrt{ts} \approx 2.2`.
# Therefore, the observed fluctuations are not significant above the background.
# Next, we will estimate the differential upper limits of the source.

###############################################################################
# Differential upper limits
# ~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In the absence of a detection, using the model directly from the fit is meaningless as its features can be
# simply due to background fluctuations.
# It is important to **set a reasonable model** on the dataset
# before proceeding with the `~gammapy.estimators.FluxPointsEstimator`. This model can come from
# measurements from other instruments, be an extrapolation of the flux
# observed at other wavelengths, come from theoretical estimations, etc.
# In particular, a model with a negative amplitude as obtained above should not be used.
#
# Note that **the computed upper limits depend on the spectral parameters of the assumed model**.
# Here, we compute the 3-sigma upper limits for assuming a spectral index of 2.0.
# We also fix the spatial parameters of the model to prevent the minimiser
# from wandering off to different regions in the FoV.
#


model1 = sky_model.copy(name="model1")
model1.parameters["amplitude"].value = 1e-14
model1.parameters["index"].value = 2.0
model1.freeze(model_type="spatial")

energy_edges = dataset.geoms["geom"].axes["energy"].edges
fpe = FluxPointsEstimator(
    selection_optional="all", energy_edges=energy_edges, n_sigma_ul=3
)

dataset.models = model1
fp1 = fpe.run(dataset)


fp1.plot(sed_type="dnde")
plt.show()


######################################################################
# Integral upper limits
# ~~~~~~~~~~~~~~~~~~~~~
#
# To compute the integral upper limits between certain energies,
# we can simply run  `~gammapy.estimators.FluxPointsEstimator`
# with one bin in energy.

emin = energy_edges[0]
emax = energy_edges[-1]
fpe2 = FluxPointsEstimator(selection_optional=["ul"], energy_edges=[emin, emax])
fp2 = fpe2.run(dataset)
print(
    f"Integral upper limit between {emin:.1f} and {emax:.1f} is {fp2.flux_ul.quantity.ravel()[0]:.2e}"
)


######################################################################
# Sensitivity estimation
# ~~~~~~~~~~~~~~~~~~~~~~
#
# We can then ask,  **would this source have been detectable given this IRF/exposure time?**
#
# The `~gammapy.estimators.FluxPointsEstimator` can be used to obtain the sensitivity,
# which can be compared to the flux prediction for a given (hypothetical) source. We have the 5-sigma
# sensitivity here, which can be configured using ``n_sigma_sensitivity``
# parameter of this estimator. Let us see what we would have seen if a Crab-like source was
# present in the center.
# Note that this computed sensitivity does not take into account the factors
# such as the minimum number of gamma-rays (see :doc:`/tutorials/analysis-1d/cta_sensitivity`)
# and is dependent on the analysis configuration.
# We compare this with the known Crab spectrum.

crab_model = create_crab_spectral_model()

fp1.flux_sensitivity.plot(label="sensitivity")
crab_model.plot(
    energy_bounds=fp1.geom.axes["energy"], sed_type="flux", label="Crab spectrum"
)
plt.grid(which="minor", alpha=0.3)

plt.legend()
plt.show()
# sphinx_gallery_thumbnail_number = 6


######################################################################
# Thus, a Crab-like source should have been above our sensitivity till around ~ 4 TeV for this specific observation.
